import inspect
import re
from collections import defaultdict
from collections.abc import Iterable, Iterator, Mapping
from dataclasses import dataclass, field
from fnmatch import fnmatch
from functools import cached_property, singledispatchmethod
from typing import TYPE_CHECKING, ClassVar, Optional

import pytest
from eth_utils import to_hex

from ape.exceptions import BlockNotFoundError, ChainError, ProviderNotConnectedError
from ape.logging import logger
from ape.pytest.utils import Scope
from ape.utils.basemodel import ManagerAccessMixin
from ape.utils.rpc import allow_disconnected

if TYPE_CHECKING:
    from ape.api.accounts import TestAccountAPI
    from ape.api.transactions import ReceiptAPI
    from ape.managers.chain import ChainManager
    from ape.managers.networks import NetworkManager
    from ape.managers.project import ProjectManager
    from ape.pytest.config import ConfigWrapper
    from ape.types.vm import SnapshotID


@dataclass()
class FixtureRebase:
    return_scope: Scope
    invalid_fixtures: dict[Scope, list[str]]


class FixtureManager(ManagerAccessMixin):
    _builtin_fixtures: ClassVar[list] = []
    _stateful_fixtures_cache: ClassVar[dict[str, bool]] = {}
    _ISOLATION_FIXTURE_REGEX = re.compile(r"_(session|package|module|class|function)_isolation")

    def __init__(self, config_wrapper: "ConfigWrapper", isolation_manager: "IsolationManager"):
        self.config_wrapper = config_wrapper
        self.isolation_manager = isolation_manager
        self._nodeid_to_fixture_map: dict[str, FixtureMap] = {}
        self._fixture_name_to_info: dict[str, dict] = {}

    @classmethod
    def set_builtins(cls, fixture_map: "FixtureMap"):
        cls._builtin_fixtures = [
            n
            for n, defs in fixture_map._arg2fixturedefs.items()
            if any("pytest" in fixture.func.__module__ for fixture in defs)
        ]

    @cached_property
    def _ape_fixtures(self) -> tuple[str, ...]:
        return tuple(
            [
                n
                for n, itm in inspect.getmembers(PytestApeFixtures)
                if callable(itm) and not n.startswith("_")
            ]
        )

    @property
    def builtin_fixtures(self) -> list[str]:
        return self._builtin_fixtures

    def is_builtin(self, name: str) -> bool:
        return name in self.builtin_fixtures

    @classmethod
    def is_isolation(cls, name: str) -> bool:
        return bool(re.match(cls._ISOLATION_FIXTURE_REGEX, name))

    def is_ape(self, name: str) -> bool:
        return name in self._ape_fixtures

    def is_custom(self, name) -> bool:
        return not self.is_builtin(name) and not self.is_ape(name) and not self.is_isolation(name)

    def get_fixtures(self, item) -> "FixtureMap":
        if isinstance(item, str):
            # Cached map referenced where only have nodeid.
            if fixture_map := self._get_cached_fixtures(item):
                return fixture_map

            raise KeyError(f"No item found with nodeid '{item}'.")

        elif fixture_map := self._get_cached_fixtures(item.nodeid):
            # Cached map.
            return fixture_map

        return self.cache_fixtures(item)

    def cache_fixtures(self, item) -> "FixtureMap":
        fixture_map = FixtureMap.from_test_item(item)
        if not FixtureManager._builtin_fixtures:
            FixtureManager.set_builtins(fixture_map)

        self._nodeid_to_fixture_map[item.nodeid] = fixture_map
        for scope, fixture_set in fixture_map.items():
            for fixture_name in fixture_set:
                if fixture_name not in self._fixture_name_to_info:
                    self._fixture_name_to_info[fixture_name] = {"scope": scope}

        return fixture_map

    def get_fixture_scope(self, fixture_name: str) -> Optional[Scope]:
        return self._fixture_name_to_info.get(fixture_name, {}).get("scope")

    def is_stateful(self, name: str) -> Optional[bool]:
        if name in self._stateful_fixtures_cache:
            # Used `@ape.fixture(chain_isolation=<bool>)
            # Or we already calculated.
            return self._stateful_fixtures_cache[name]

        try:
            is_auto_mine = self.provider.auto_mine
        except (NotImplementedError, ProviderNotConnectedError):
            # Assume it's on since it can't be turned off.
            is_auto_mine = True
        except AttributeError:
            # If not a test-provider, we don't know.
            return None

        if not is_auto_mine:
            # When auto-mine is disabled, it's unknown.
            return None

        elif not (info := self._fixture_name_to_info.get(name)):
            # Statefulness not yet tracked. Unknown.
            return None

        setup_block = info.get("setup_block")
        teardown_block = info.get("teardown_block")
        if setup_block is None or teardown_block is None:
            # Blocks no set. Unknown.
            return None

        # If the two are not equal, state has changed.
        is_stateful = setup_block != teardown_block
        self._stateful_fixtures_cache[name] = is_stateful

        # Clear out blocks since they are no longer needed.
        self._fixture_name_to_info[name] = {
            k: v
            for k, v in self._fixture_name_to_info[name].items()
            if k not in ("setup_block", "teardown_block")
        }

        return is_stateful

    def add_fixture_info(self, name: str, **info):
        if name not in self._fixture_name_to_info:
            self._fixture_name_to_info[name] = info
        else:
            self._fixture_name_to_info[name] = {
                **self._fixture_name_to_info[name],
                **info,
            }

    def _get_cached_fixtures(self, nodeid: str) -> Optional["FixtureMap"]:
        return self._nodeid_to_fixture_map.get(nodeid)

    def needs_rebase(self, new_fixtures: list[str], snapshot: "Snapshot") -> bool:
        return self.config_wrapper.enable_fixture_rebasing and bool(
            new_fixtures and snapshot.fixtures
        )

    def rebase(self, scope: Scope, fixtures: "FixtureMap"):
        if not (rebase := self._get_rebase(scope)):
            # Rebase avoided: nothing would change.
            return

        from ape.pytest.warnings import warn_invalid_isolation

        warn_invalid_isolation()
        self.isolation_manager.restore(rebase.return_scope)

        # Invalidate fixtures by clearing out their cached result.
        invalidated = []
        for invalid_scope, invalid_fixture_ls in rebase.invalid_fixtures.items():
            for invalid_fixture in invalid_fixture_ls:
                info_ls = fixtures.get_info(invalid_fixture)
                for info in info_ls:
                    if self.is_stateful(info.argname) is False:
                        # It has been determined that this fixture is not stateful.
                        continue

                    info.cached_result = None
                    invalidated.append(info.argname)

            # Also, invalidate the corresponding isolation fixture.
            if invalid_isolation_fixture_ls := fixtures.get_info(
                invalid_scope.isolation_fixturename
            ):
                for invalid_isolation_fixture in invalid_isolation_fixture_ls:
                    invalid_isolation_fixture.cached_result = None
                    invalidated.append(invalid_isolation_fixture.argname)

            if invalidated and self.config_wrapper.verbosity:
                log = "rebase"
                if rebase.return_scope is not None:
                    log = f"{log} scope={rebase.return_scope}"

    def _get_rebase(self, scope: Scope) -> Optional[FixtureRebase]:
        # Check for fixtures that are now invalid. For example, imagine a session
        # fixture comes into play after the module snapshot has been set.
        # Once we restore the module's state and move to the next module,
        # that session fixture will no longer exist. To remedy this situation,
        # we invalidate the lower-scoped fixtures and re-snapshot everything.
        scope_to_revert = None
        invalids = defaultdict(list)
        for next_snapshot in self.isolation_manager.next_snapshots(scope):
            if next_snapshot.identifier is None:
                # Thankfully, we haven't reached this scope yet (or it is disabled).
                # In this case, things are running in a correct/performant order.
                continue

            if scope_to_revert is None:
                # Revert to the closest scope to use. For example, a new
                # session comes in, but we have already calculated a module
                # and a class, revert to pre-module and invalidate the module
                # and class fixtures.
                scope_to_revert = next_snapshot.scope

            # All stateful fixtures downward are "below scope"
            fixtures = [f for f in next_snapshot.fixtures if self.is_stateful(f) is not False]
            invalids[next_snapshot.scope].extend(fixtures)

        invalids_dict = dict(invalids)
        return (
            FixtureRebase(return_scope=scope_to_revert, invalid_fixtures=invalids_dict)
            if scope_to_revert is not None and any(len(ls) > 0 for ls in invalids_dict.values())
            else None
        )


class FixtureMap(dict[Scope, list[str]]):
    def __init__(self, item):
        self._item = item
        self._parametrized_names: Optional[list[str]] = None
        super().__init__(
            {
                Scope.SESSION: [],
                Scope.PACKAGE: [],
                Scope.MODULE: [],
                Scope.CLASS: [],
                Scope.FUNCTION: [],
            }
        )

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self._item.nodeid}>"

    @classmethod
    def from_test_item(cls, item) -> "FixtureMap":
        obj = cls(item)
        for name, info_ls in obj._arg2fixturedefs.items():
            if not info_ls or name not in item.fixturenames:
                continue

            for info in info_ls:
                obj[info.scope].append(name)

        return obj

    @property
    def names(self) -> list[str]:
        """
        Outputs in correct order for item.fixturenames.
        Also, injects isolation fixtures if needed.
        """
        result = []
        for scope, ls in self.items():
            # NOTE: For function scoped, we always add the isolation fixture.
            if not ls and scope is not Scope.FUNCTION:
                continue

            result.append(scope.isolation_fixturename)
            result.extend(ls)

        return result

    @property
    def parameters(self) -> list[str]:
        """
        Test-parameters (not fixtures!)
        """
        return [n for n in self._item.fixturenames if n not in self._arg2fixturedefs]

    @property
    def isolation(self) -> list[str]:
        return [n.lstrip("_").split("_")[0] for n in self.names if FixtureManager.is_isolation(n)]

    @property
    def parametrized(self) -> dict[str, list]:
        if self._parametrized_names is not None:
            # We have already done this.
            return {
                n: ls for n, ls in self._arg2fixturedefs.items() if n in self._parametrized_names
            }

        # Calculating for first time.
        self._parametrized_names = []
        result: dict[str, list] = {}
        for name, info_ls in self._arg2fixturedefs.items():
            if name not in self._item.fixturenames or not any(info.params for info in info_ls):
                continue

            self._parametrized_names.append(name)
            result[name] = info_ls

        return result

    @property
    def _arg2fixturedefs(self) -> Mapping:
        return self._item.session._fixturemanager._arg2fixturedefs

    @singledispatchmethod
    def __setitem__(self, key, value):
        raise NotImplementedError(type(key))

    @__setitem__.register
    def __setitem_int(self, key: int, value: list[str]):
        super().__setitem__(Scope(key), value)

    @__setitem__.register
    def __setitem_str(self, key: str, value: list[str]):
        for scope in Scope:
            if f"{scope}" == key:
                super().__setitem__(scope, value)
                return

        raise KeyError(key)

    @__setitem__.register
    def __setitem_scope(self, key: Scope, value: list[str]):
        super().__setitem__(key, value)

    @singledispatchmethod
    def __getitem__(self, key):
        # NOTE: Not using singledispatchmethod because it requires
        #  types at runtime.
        if isinstance(key, Scope):
            return self.__getitem_scope(key)
        elif isinstance(key, str):
            return self.__getitem_str(key)
        elif isinstance(key, int):
            return self.__getitem_int(key)

        raise NotImplementedError(type(key))

    @__getitem__.register
    def __getitem_int(self, key: int) -> list[str]:
        return super().__getitem__(Scope(key))

    @__getitem__.register
    def __getitem_str(self, key: str) -> list[str]:
        for scope in Scope:
            if f"{scope}" == key:
                return super().__getitem__(scope)

        raise KeyError(key)

    @__getitem__.register
    def __getitem_scope(self, key: Scope) -> list[str]:
        return super().__getitem__(key)

    def get_info(self, name: str) -> list:
        """
        Get fixture info.

        Args:
            name (str):

        Returns:
            list of info
        """
        if name not in self._arg2fixturedefs:
            return []

        return self._arg2fixturedefs[name]

    def is_known(self, name: str) -> bool:
        """
        True when fixture-info is known for the given fixture name.
        """
        return name in self._arg2fixturedefs

    def is_iterating(self, name: str) -> bool:
        """
        True when is a non-function scoped parametrized fixture that hasn't
        fully iterated.
        """
        if name not in self.parametrized:
            return False

        elif not (info_ls := self.get_info(name)):
            return False

        for info in info_ls:
            if not info.params:
                continue

            if not info.cached_result:
                return True  # First iteration

            elif len(info.cached_result) < 2:
                continue  # ?

            last_param_ran = info.cached_result[1]
            last_param = info.params[-1]
            if last_param_ran != last_param:
                return True  # Is iterating.

        return False

    def apply_fixturenames(self):
        """
        Set the fixturenames on the test item in the order they should be used.
        Carefully ignore non-fixtures, such as keys from parametrized tests.
        """
        self._item.fixturenames = [*self.names, *self.parameters]


class PytestApeFixtures(ManagerAccessMixin):
    # NOTE: Avoid including links, markdown, or rst in method-docs
    # for fixtures, as they are used in output from the command
    # `ape test -q --fixture` (`pytest -q --fixture`).

    def __init__(self, config_wrapper: "ConfigWrapper", isolation_manager: "IsolationManager"):
        self.config_wrapper = config_wrapper
        self.isolation_manager = isolation_manager

    @pytest.fixture(scope="session")
    def accounts(self) -> list["TestAccountAPI"]:
        """
        A collection of pre-funded accounts.
        """
        return self.account_manager.test_accounts

    @pytest.fixture(scope="session")
    def compilers(self):
        """
        Access compiler manager directly.
        """
        return self.compiler_manager

    @pytest.fixture(scope="session")
    def chain(self) -> "ChainManager":
        """
        Manipulate the blockchain, such as mine or change the pending timestamp.
        """
        return self.chain_manager

    @pytest.fixture(scope="session")
    def networks(self) -> "NetworkManager":
        """
        Connect to other networks in your tests.
        """
        return self.network_manager

    @pytest.fixture(scope="session")
    def project(self) -> "ProjectManager":
        """
        Access contract types and dependencies.
        """
        return self.local_project

    @pytest.fixture(scope="session")
    def Contract(self):
        """
        Instantiate a reference to an on-chain contract
        using its address (same as ``ape.Contract``).
        """
        return self.chain_manager.contracts.instance_at

    @pytest.fixture(scope="session")
    def _session_isolation(self) -> Iterator[None]:
        yield from self.isolation_manager.isolation(Scope.SESSION)

    @pytest.fixture(scope="package")
    def _package_isolation(self) -> Iterator[None]:
        yield from self.isolation_manager.isolation(Scope.PACKAGE)

    @pytest.fixture(scope="module")
    def _module_isolation(self) -> Iterator[None]:
        yield from self.isolation_manager.isolation(Scope.MODULE)

    @pytest.fixture(scope="class")
    def _class_isolation(self) -> Iterator[None]:
        yield from self.isolation_manager.isolation(Scope.CLASS)

    @pytest.fixture(scope="function")
    def _function_isolation(self) -> Iterator[None]:
        yield from self.isolation_manager.isolation(Scope.FUNCTION)


@dataclass
class Snapshot:
    """
    All the data necessary for accurately supporting isolation.
    """

    scope: Scope
    """Corresponds to fixture scope."""

    identifier: Optional["SnapshotID"] = None
    """Snapshot ID taken before the peer-fixtures in the same scope."""

    fixtures: list = field(default_factory=list)
    """All peer fixtures, tracked so we know when new ones are added."""

    def append_fixtures(self, fixtures: Iterable[str]):
        for fixture in fixtures:
            if fixture in self.fixtures:
                continue

            self.fixtures.append(fixture)


class SnapshotRegistry(dict[Scope, Snapshot]):
    def __init__(self):
        super().__init__(
            {
                Scope.SESSION: Snapshot(Scope.SESSION),
                Scope.PACKAGE: Snapshot(Scope.PACKAGE),
                Scope.MODULE: Snapshot(Scope.MODULE),
                Scope.CLASS: Snapshot(Scope.CLASS),
                Scope.FUNCTION: Snapshot(Scope.FUNCTION),
            }
        )

    def get_snapshot_id(self, scope: Scope) -> Optional["SnapshotID"]:
        return self[scope].identifier

    def set_snapshot_id(self, scope: Scope, snapshot_id: "SnapshotID"):
        self[scope].identifier = snapshot_id

    def clear_snapshot_id(self, scope: Scope):
        self[scope].identifier = None

    def next_snapshots(self, scope: Scope) -> Iterator[Snapshot]:
        for scope_value in range(scope + 1, Scope.FUNCTION + 1):
            yield self[scope_value]  # type: ignore

    def extend_fixtures(self, scope: Scope, fixtures: Iterable[str]):
        self[scope].fixtures.extend(fixtures)


class IsolationManager(ManagerAccessMixin):
    supported: bool = True
    snapshots: SnapshotRegistry = SnapshotRegistry()

    def __init__(
        self,
        config_wrapper: "ConfigWrapper",
        receipt_capture: "ReceiptCapture",
        chain_snapshots: Optional[dict] = None,
    ):
        self.config_wrapper = config_wrapper
        self.receipt_capture = receipt_capture
        self._chain_snapshots = chain_snapshots

    @cached_property
    def _track_transactions(self) -> bool:
        return (
            self.network_manager.provider is not None
            and self.provider.is_connected
            and (self.config_wrapper.track_gas or self.config_wrapper.track_coverage)
        )

    @property
    def chain_snapshots(self) -> dict:
        return self._chain_snapshots or self.chain_manager._snapshots

    def get_snapshot(self, scope: Scope) -> Snapshot:
        return self.snapshots[scope]

    def extend_fixtures(self, scope: Scope, fixtures: Iterable[str]):
        self.snapshots.extend_fixtures(scope, fixtures)

    def next_snapshots(self, scope: Scope) -> Iterator[Snapshot]:
        yield from self.snapshots.next_snapshots(scope)

    def isolation(self, scope: Scope) -> Iterator[None]:
        """
        Isolation logic used to implement isolation fixtures for each pytest scope.
        When tracing support is available, will also assist in capturing receipts.
        """
        self.set_snapshot(scope)
        if self._track_transactions:
            did_yield = False
            try:
                with self.receipt_capture:
                    yield
                    did_yield = True

            except BlockNotFoundError:
                if not did_yield:
                    # Prevent double yielding.
                    yield
        else:
            yield

        self.restore(scope)

    def set_snapshot(self, scope: Scope):
        # Also can be used to re-set snapshot.
        if not self.supported or not self.config_wrapper.get_isolation(scope):
            return

        try:
            snapshot_id = self.take_snapshot()
        except Exception:
            self.supported = False
        else:
            if snapshot_id is not None:
                self.snapshots.set_snapshot_id(scope, snapshot_id)

    @allow_disconnected
    def take_snapshot(self) -> Optional["SnapshotID"]:
        try:
            return self.chain_manager.snapshot()
        except NotImplementedError:
            logger.warning(
                "The connected provider does not support snapshotting. "
                "Tests will not be completely isolated."
            )
            # To avoid trying again
            self.supported = False

        return None

    @allow_disconnected
    def restore(self, scope: Scope):
        # NOTE: self._supported may have gotten set to False
        #   someplace else _after_ snapshotting succeeded.
        if not self.supported or not self.config_wrapper.get_isolation(scope):
            return

        self.restore_snapshot(scope)

    @allow_disconnected
    def restore_snapshot(self, scope: Scope):
        snapshot_id = self.snapshots.get_snapshot_id(scope)
        if snapshot_id is None:
            return

        elif snapshot_id not in self.chain_snapshots[self.provider.chain_id]:
            # Still clear out.
            self.snapshots.clear_snapshot_id(scope)
            return

        try:
            self._restore(snapshot_id)
        except NotImplementedError:
            logger.warning(
                "The connected provider does not support snapshotting. "
                "Tests will not be completely isolated."
            )
            # To avoid trying again
            self.supported = False

        self.snapshots.clear_snapshot_id(scope)

    def _restore(self, snapshot_id: "SnapshotID"):
        self.chain_manager.restore(snapshot_id)


class ReceiptCapture(ManagerAccessMixin):
    receipt_map: dict[str, dict[str, "ReceiptAPI"]] = {}
    enter_blocks: list[int] = []

    def __init__(self, config_wrapper: "ConfigWrapper"):
        self.config_wrapper = config_wrapper

    def __enter__(self):
        block_number = self._get_block_number()
        if block_number is not None:
            self.enter_blocks.append(block_number)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.enter_blocks:
            return

        start_block = self.enter_blocks.pop()
        stop_block = self._get_block_number()
        if stop_block is None or start_block > stop_block:
            return

        self.capture_range(start_block, stop_block)

    def capture_range(self, start_block: int, stop_block: int):
        blocks = self.chain_manager.blocks.range(start_block, stop_block + 1)
        transactions = [t for b in blocks for t in b.transactions]

        for txn in transactions:
            try:
                txn_hash = to_hex(txn.txn_hash)
            except Exception:
                # Might have been from an impersonated account.
                # Those txns need to be added separately, same as tracing calls.
                # Likely, it was already accounted before this point.
                continue

            self.capture(txn_hash)

    def capture(self, transaction_hash: str):
        try:
            receipt = self.chain_manager.history[transaction_hash]
        except ChainError:
            return

        if not receipt:
            return

        elif not (contract_address := (receipt.receiver or receipt.contract_address)):
            return

        elif not (contract_type := self.chain_manager.contracts.get(contract_address)):
            # Not an invoke-transaction or a known address
            return

        elif not (source_id := (contract_type.source_id or None)):
            # Not a local or known contract type.
            return

        elif source_id not in self.receipt_map:
            self.receipt_map[source_id] = {}

        if transaction_hash in self.receipt_map[source_id]:
            # Transaction already known.
            return

        self.receipt_map[source_id][transaction_hash] = receipt
        if self.config_wrapper.track_gas:
            receipt.track_gas()

        if self.config_wrapper.track_coverage:
            receipt.track_coverage()

    def clear(self):
        self.receipt_map = {}
        self.enter_blocks = []

    @allow_disconnected
    def _get_block_number(self) -> Optional[int]:
        return self.provider.get_block("latest").number

    def _exclude_from_gas_report(
        self, contract_name: str, method_name: Optional[str] = None
    ) -> bool:
        """
        Helper method to determine if a certain contract / method combination should be
        excluded from the gas report.
        """
        for exclusion in self.config_wrapper.gas_exclusions:
            # Default to looking at all contracts
            contract_pattern = exclusion.contract_name
            if not fnmatch(contract_name, contract_pattern) or not method_name:
                continue

            method_pattern = exclusion.method_name
            if not method_pattern or fnmatch(method_name, method_pattern):
                return True

        return False


def fixture(chain_isolation: Optional[bool], **kwargs):
    """
    A thin-wrapper around ``@pytest.fixture`` with extra capabilities.
    Set ``chain_isolation`` to ``False`` to signal to Ape that this fixture's
    cached result is the same regardless of block number and it does not
    need to be invalidated during times or pytest-scoped based chain rebasing.

    Usage example::

        import ape
        from ape_tokens import tokens


        @ape.fixture(scope="session", chain_isolation=False, params=("WETH", "DAI", "BAT"))
        def token_addresses(request):
            return tokens[request].address

    """

    def decorator(fixture_function):
        if chain_isolation is not None:
            name = kwargs.get("name", fixture_function.__name__)
            FixtureManager._stateful_fixtures_cache[name] = chain_isolation

        return pytest.fixture(fixture_function, **kwargs)

    return decorator



import json
from collections.abc import Collection
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Generic, Optional, TypeVar, Union

from ethpm_types import ABI, ContractType
from pydantic import BaseModel

from ape.api.networks import ProxyInfoAPI
from ape.api.query import ContractCreation, ContractCreationQuery
from ape.contracts.base import ContractContainer, ContractInstance
from ape.exceptions import ApeException, ContractNotFoundError, ConversionError, CustomError
from ape.logging import logger
from ape.managers._deploymentscache import Deployment, DeploymentDiskCache
from ape.managers.base import BaseManager
from ape.types.address import AddressType
from ape.utils.misc import nonreentrant
from ape.utils.os import CacheDirectory

if TYPE_CHECKING:
    from eth_pydantic_types import HexBytes

    from ape.api.transactions import ReceiptAPI


_BASE_MODEL = TypeVar("_BASE_MODEL", bound=BaseModel)


class ApeDataCache(CacheDirectory, Generic[_BASE_MODEL]):
    """
    A wrapper around some cached models in the data directory,
    such as the cached contract types.
    """

    def __init__(
        self,
        base_data_folder: Path,
        ecosystem_key: str,
        network_key: str,
        key: str,
        model_type: type[_BASE_MODEL],
    ):
        data_folder = base_data_folder / ecosystem_key
        base_path = data_folder / network_key
        self._model_type = model_type
        self.memory: dict[str, _BASE_MODEL] = {}

        # Only write if we are not testing!
        self._write_to_disk = not network_key.endswith("-fork") and network_key != "local"
        # Read from disk if using forks or live networks.
        self._read_from_disk = network_key.endswith("-fork") or network_key != "local"

        super().__init__(base_path / key)

    def __getitem__(self, key: str) -> Optional[_BASE_MODEL]:  # type: ignore
        return self.get_type(key)

    def __setitem__(self, key: str, value: _BASE_MODEL):  # type: ignore
        self.memory[key] = value
        if self._write_to_disk:
            # Cache to disk.
            self.cache_data(key, value.model_dump(mode="json"))

    def __delitem__(self, key: str):
        self.memory.pop(key, None)
        if self._write_to_disk:
            # Delete the cache file.
            self.delete_data(key)

    def __contains__(self, key: str) -> bool:
        try:
            return bool(self[key])
        except KeyError:
            return False

    def get_type(self, key: str) -> Optional[_BASE_MODEL]:
        if model := self.memory.get(key):
            return model

        elif self._read_from_disk:
            if data := self.get_data(key):
                # Found on disk.
                model = self._model_type.model_validate(data)
                # Cache locally for next time.
                self.memory[key] = model
                return model

        return None


class ContractCache(BaseManager):
    """
    A collection of cached contracts. Contracts can be cached in two ways:

    1. An in-memory cache of locally deployed contracts
    2. A cache of contracts per network (only permanent networks are stored this way)

    When retrieving a contract, if a :class:`~ape.api.explorers.ExplorerAPI` is used,
    it will be cached to disk for faster look-up next time.
    """

    # ecosystem_name -> network_name -> cache_name -> cache
    _caches: dict[str, dict[str, dict[str, ApeDataCache]]] = {}

    # chain_id -> address -> custom_err
    # Cached to prevent calling `new_class` multiple times with conflicts.
    _custom_error_types: dict[int, dict[AddressType, set[type[CustomError]]]] = {}

    @property
    def contract_types(self) -> ApeDataCache[ContractType]:
        return self._get_data_cache("contract_types", ContractType)

    @property
    def proxy_infos(self) -> ApeDataCache[ProxyInfoAPI]:
        return self._get_data_cache("proxy_info", ProxyInfoAPI)

    @property
    def blueprints(self) -> ApeDataCache[ContractType]:
        return self._get_data_cache("blueprints", ContractType)

    @property
    def contract_creations(self) -> ApeDataCache[ContractCreation]:
        return self._get_data_cache("contract_creation", ContractCreation)

    def _get_data_cache(
        self,
        key: str,
        model_type: type,
        ecosystem_key: Optional[str] = None,
        network_key: Optional[str] = None,
    ):
        ecosystem_name = ecosystem_key or self.provider.network.ecosystem.name
        network_name = network_key or self.provider.network.name.replace("-fork", "")
        self._caches.setdefault(ecosystem_name, {})
        self._caches[ecosystem_name].setdefault(network_name, {})

        if cache := self._caches[ecosystem_name][network_name].get(key):
            return cache

        self._caches[ecosystem_name][network_name][key] = ApeDataCache(
            self.config_manager.DATA_FOLDER, ecosystem_name, network_name, key, model_type
        )
        return self._caches[ecosystem_name][network_name][key]

    @cached_property
    def deployments(self) -> DeploymentDiskCache:
        """A manager for contract deployments across networks."""
        return DeploymentDiskCache()

    def __setitem__(
        self, address: AddressType, item: Union[ContractType, ProxyInfoAPI, ContractCreation]
    ):
        """
        Cache the given contract type. Contracts are cached in memory per session.
        In live networks, contracts also get cached to disk at
        ``.ape/{ecosystem_name}/{network_name}/contract_types/{address}.json``
        for faster look-up next time.

        Args:
            address (AddressType): The on-chain address of the contract.
            item (ContractType | ProxyInfoAPI | ContractCreation): The contract's type, proxy info,
              or creation metadata.
        """
        # Note: Can't cache blueprints this way.
        address = self.provider.network.ecosystem.decode_address(int(address, 16))
        if isinstance(item, ContractType):
            self.cache_contract_type(address, item)
        elif isinstance(item, ProxyInfoAPI):
            self.cache_proxy_info(address, item)
        elif isinstance(item, ContractCreation):
            self.cache_contract_creation(address, item)
        elif contract_type := getattr(item, "contract_type", None):
            self.cache_contract_type(address, contract_type)
        else:
            raise TypeError(item)

    def cache_contract_type(
        self,
        address: AddressType,
        contract_type: ContractType,
        ecosystem_key: Optional[str] = None,
        network_key: Optional[str] = None,
    ):
        """
        Cache a contract type at the given address for the given network.
        If not connected, you must provider both ``ecosystem_key:`` and
        ``network_key::``.

        Args:
            address (AddressType): The address key.
            contract_type (ContractType): The contract type to cache.
            ecosystem_key (str | None): The ecosystem key. Defaults to
              the connected ecosystem's name.
            network_key (str | None): The network key. Defaults to the
              connected network's name.
        """
        # Get the cache in a way that doesn't require an active connection.
        cache = self._get_data_cache(
            "contract_types", ContractType, ecosystem_key=ecosystem_key, network_key=network_key
        )
        cache[address] = contract_type

        # NOTE: The txn_hash is not included when caching this way.
        if name := contract_type.name:
            self.deployments.cache_deployment(
                address, name, ecosystem_key=ecosystem_key, network_key=network_key
            )

    def cache_contract_creation(
        self,
        address: AddressType,
        contract_creation: ContractCreation,
        ecosystem_key: Optional[str] = None,
        network_key: Optional[str] = None,
    ):
        """
        Cache a contract creation object.

        Args:
            address (AddressType): The address of the contract.
            contract_creation (ContractCreation): The object to cache.
            ecosystem_key (str | None): The ecosystem key. Defaults to
              the connected ecosystem's name.
            network_key (str | None): The network key. Defaults to the
              connected network's name.
        """
        # Get the cache in a way that doesn't require an active connection.
        cache = self._get_data_cache(
            "contract_creation",
            ContractCreation,
            ecosystem_key=ecosystem_key,
            network_key=network_key,
        )
        cache[address] = contract_creation

    def __delitem__(self, address: AddressType):
        """
        Delete a cached contract.
        If using a live network, it will also delete the file-cache for the contract.

        Args:
            address (AddressType): The address to remove from the cache.
        """
        del self.contract_types[address]
        self._delete_proxy(address)
        del self.contract_creations[address]

    @contextmanager
    def use_temporary_caches(self):
        """
        Create temporary context where there are no cached items.
        Useful for testing.
        """
        caches = self._caches
        self._caches = {}
        with self.deployments.use_temporary_cache():
            yield

        self._caches = caches

    def _delete_proxy(self, address: AddressType):
        if info := self.proxy_infos[address]:
            target = info.target
            del self.proxy_infos[target]
            del self.contract_types[target]

    def __contains__(self, address: AddressType) -> bool:
        return self.get(address) is not None

    def cache_deployment(
        self,
        contract_instance: ContractInstance,
        proxy_info: Optional[ProxyInfoAPI] = None,
        detect_proxy: bool = True,
    ):
        """
        Cache the given contract instance's type and deployment information.

        Args:
            contract_instance (:class:`~ape.contracts.base.ContractInstance`): The contract
              to cache.
            proxy_info (Optional[ProxyInfoAPI]): Pass in the proxy info, if it is known, to
              avoid the potentially expensive look-up.
            detect_proxy (bool): Set to ``False`` to avoid detecting if the contract is a
              proxy.
        """
        address = contract_instance.address
        contract_type = contract_instance.contract_type  # may be a proxy

        # Cache contract type in memory before proxy check,
        # in case it is needed somewhere. It may get overridden.
        self.contract_types.memory[address] = contract_type

        if proxy_info:
            # Was given proxy info.
            self._cache_proxy_contract(address, proxy_info, contract_type, contract_instance)

        elif detect_proxy:
            # Proxy info was not provided. Use the connected ecosystem to figure it out.
            if proxy_info := self.provider.network.ecosystem.get_proxy_info(address):
                # The user is caching a deployment of a proxy with the target already set.
                self._cache_proxy_contract(address, proxy_info, contract_type, contract_instance)

            else:
                # Cache as normal.
                self.contract_types[address] = contract_type

        else:
            # Cache as normal; do not do expensive proxy detection.
            self.contract_types[address] = contract_type

        # Cache the deployment now.
        txn_hash = contract_instance.txn_hash
        if contract_name := contract_type.name:
            self.deployments.cache_deployment(address, contract_name, transaction_hash=txn_hash)

        return contract_type

    def _cache_proxy_contract(
        self,
        address: AddressType,
        proxy_info: ProxyInfoAPI,
        contract_type: ContractType,
        contract_instance: ContractInstance,
    ):
        self.cache_proxy_info(address, proxy_info)
        if implementation_contract := self.get(proxy_info.target):
            updated_proxy_contract = _get_combined_contract_type(
                contract_type, proxy_info, implementation_contract
            )
            self.contract_types[address] = updated_proxy_contract

            # Use this contract type in the user's contract instance.
            contract_instance.contract_type = updated_proxy_contract
        else:
            # No implementation yet. Just cache proxy.
            self.contract_types[address] = contract_type

    def cache_proxy_info(self, address: AddressType, proxy_info: ProxyInfoAPI):
        """
        Cache proxy info for a particular address, useful for plugins adding already
        deployed proxies. When you deploy a proxy locally, it will also call this method.

        Args:
            address (AddressType): The address of the proxy contract.
            proxy_info (:class:`~ape.api.networks.ProxyInfo`): The proxy info class
              to cache.
        """
        self.proxy_infos[address] = proxy_info

    def cache_blueprint(self, blueprint_id: str, contract_type: ContractType):
        """
        Cache a contract blueprint.

        Args:
            blueprint_id (``str``): The ID of the blueprint. For example, in EIP-5202,
              it would be the address of the deployed blueprint. For Starknet, it would
              be the class identifier.
            contract_type (``ContractType``): The contract type associated with the blueprint.
        """
        self.blueprints[blueprint_id] = contract_type

    def get_proxy_info(self, address: AddressType) -> Optional[ProxyInfoAPI]:
        """
        Get proxy information about a contract using its address,
        either from a local cache, a disk cache, or the provider.

        Args:
            address (AddressType): The address of the proxy contract.

        Returns:
            Optional[:class:`~ape.api.networks.ProxyInfoAPI`]
        """
        return self.proxy_infos[address]

    def get_creation_metadata(self, address: AddressType) -> Optional[ContractCreation]:
        """
        Get contract creation metadata containing txn_hash, deployer, factory, block.

        Args:
            address (AddressType): The address of the contract.

        Returns:
            Optional[:class:`~ape.api.query.ContractCreation`]
        """
        if creation := self.contract_creations[address]:
            return creation

        # Query and cache.
        query = ContractCreationQuery(columns=["*"], contract=address)
        get_creation = self.query_manager.query(query)

        try:
            if not (creation := next(get_creation, None)):  # type: ignore[arg-type]
                return None

        except ApeException:
            return None

        self.contract_creations[address] = creation
        return creation

    def get_blueprint(self, blueprint_id: str) -> Optional[ContractType]:
        """
        Get a cached blueprint contract type.

        Args:
            blueprint_id (``str``): The unique identifier used when caching
              the blueprint.

        Returns:
            ``ContractType``
        """
        return self.blueprints[blueprint_id]

    def _get_errors(
        self, address: AddressType, chain_id: Optional[int] = None
    ) -> set[type[CustomError]]:
        if chain_id is None and self.network_manager.active_provider is not None:
            chain_id = self.provider.chain_id
        elif chain_id is None:
            raise ValueError("Missing chain ID.")

        if chain_id not in self._custom_error_types:
            return set()

        errors = self._custom_error_types[chain_id]
        if address in errors:
            return errors[address]

        return set()

    def _cache_error(
        self, address: AddressType, error: type[CustomError], chain_id: Optional[int] = None
    ):
        if chain_id is None and self.network_manager.active_provider is not None:
            chain_id = self.provider.chain_id
        elif chain_id is None:
            raise ValueError("Missing chain ID.")

        if chain_id not in self._custom_error_types:
            self._custom_error_types[chain_id] = {address: set()}
        elif address not in self._custom_error_types[chain_id]:
            self._custom_error_types[chain_id][address] = set()

        self._custom_error_types[chain_id][address].add(error)

    def __getitem__(self, address: AddressType) -> ContractType:
        contract_type = self.get(address)
        if not contract_type:
            # Create error message from custom exception cls.
            err = ContractNotFoundError(
                address, self.provider.network.explorer is not None, self.provider.network_choice
            )
            # Must raise KeyError.
            raise KeyError(str(err))

        return contract_type

    def get_multiple(
        self, addresses: Collection[AddressType], concurrency: Optional[int] = None
    ) -> dict[AddressType, ContractType]:
        """
        Get contract types for all given addresses.

        Args:
            addresses (list[AddressType): A list of addresses to get contract types for.
            concurrency (Optional[int]): The number of threads to use. Defaults to
              ``min(4, len(addresses))``.

        Returns:
            dict[AddressType, ContractType]: A mapping of addresses to their respective
            contract types.
        """
        if not addresses:
            logger.warning("No addresses provided.")
            return {}

        def get_contract_type(addr: AddressType):
            addr = self.conversion_manager.convert(addr, AddressType)
            ct = self.get(addr)

            if not ct:
                logger.warning(f"Failed to locate contract at '{addr}'.")
                return addr, None
            else:
                return addr, ct

        converted_addresses: list[AddressType] = []
        for address in converted_addresses:
            if not self.conversion_manager.is_type(address, AddressType):
                converted_address = self.conversion_manager.convert(address, AddressType)
                converted_addresses.append(converted_address)
            else:
                converted_addresses.append(address)

        contract_types = {}
        default_max_threads = 4
        max_threads = (
            concurrency
            if concurrency is not None
            else min(len(addresses), default_max_threads) or default_max_threads
        )
        with ThreadPoolExecutor(max_workers=max_threads) as pool:
            for address, contract_type in pool.map(get_contract_type, addresses):
                if contract_type is None:
                    continue

                contract_types[address] = contract_type

        return contract_types

    @nonreentrant(key_fn=lambda *args, **kwargs: args[1])
    def get(
        self,
        address: AddressType,
        default: Optional[ContractType] = None,
        fetch_from_explorer: bool = True,
        proxy_info: Optional[ProxyInfoAPI] = None,
        detect_proxy: bool = True,
    ) -> Optional[ContractType]:
        """
        Get a contract type by address.
        If the contract is cached, it will return the contract from the cache.
        Otherwise, if on a live network, it fetches it from the
        :class:`~ape.api.explorers.ExplorerAPI`.

        Args:
            address (AddressType): The address of the contract.
            default (Optional[ContractType]): A default contract when none is found.
              Defaults to ``None``.
            fetch_from_explorer (bool): Set to ``False`` to avoid fetching from an
              explorer. Defaults to ``True``. Only fetches if it needs to (uses disk
              & memory caching otherwise).
            proxy_info (Optional[ProxyInfoAPI]): Pass in the proxy info, if it is known,
              to avoid the potentially expensive look-up.
            detect_proxy (bool): Set to ``False`` to avoid detecting if it is a proxy.

        Returns:
            Optional[ContractType]: The contract type if it was able to get one,
              otherwise the default parameter.
        """
        try:
            address_key: AddressType = self.conversion_manager.convert(address, AddressType)
        except ConversionError:
            if not address.startswith("0x"):
                # Still raise conversion errors for ENS and such.
                raise

            # In this case, it at least _looked_ like an address.
            return None

        if contract_type := self.contract_types[address_key]:
            # The ContractType was previously cached.
            if default and default != contract_type:
                # The given default ContractType is different than the cached one.
                # Merge the two and cache the merged result.
                combined_contract_type = _merge_contract_types(contract_type, default)
                self.contract_types[address_key] = combined_contract_type
                return combined_contract_type

            return contract_type

        # Contract is not cached yet. Check broader sources, such as an explorer.
        if not proxy_info and detect_proxy:
            # Proxy info not provided. Attempt to detect.
            if not (proxy_info := self.proxy_infos[address_key]):
                if proxy_info := self.provider.network.ecosystem.get_proxy_info(address_key):
                    self.proxy_infos[address_key] = proxy_info

        if proxy_info:
            if proxy_contract_type := self._get_proxy_contract_type(
                address_key,
                proxy_info,
                fetch_from_explorer=fetch_from_explorer,
                default=default,
            ):
                # `proxy_contract_type` is one of the following:
                # 1. A ContractType with the combined proxy and implementation ABIs
                # 2. Implementation-only ABI ContractType (like forwarder proxies)
                # 3. Proxy only ABI (e.g. unverified implementation ContractType)
                return proxy_contract_type

        # Also gets cached to disk for faster lookup next time.
        if fetch_from_explorer:
            contract_type = self._get_contract_type_from_explorer(address_key)

        # Cache locally for faster in-session look-up.
        if contract_type:
            self.contract_types[address_key] = contract_type
        elif default:
            self.contract_types[address_key] = default
            return default

        return contract_type

    def _get_proxy_contract_type(
        self,
        address: AddressType,
        proxy_info: ProxyInfoAPI,
        fetch_from_explorer: bool = True,
        default: Optional[ContractType] = None,
    ) -> Optional[ContractType]:
        """
        Combines the discoverable ABIs from the proxy contract and its implementation.
        """
        implementation_contract_type = self._get_contract_type(
            proxy_info.target,
            fetch_from_explorer=fetch_from_explorer,
            default=default,
        )
        proxy_contract_type = self._get_contract_type(
            address, fetch_from_explorer=fetch_from_explorer
        )
        if proxy_contract_type is not None and implementation_contract_type is not None:
            combined_contract = _get_combined_contract_type(
                proxy_contract_type, proxy_info, implementation_contract_type
            )
            self.contract_types[address] = combined_contract
            return combined_contract

        elif implementation_contract_type is not None:
            contract_type_to_cache = implementation_contract_type
            self.contract_types[address] = implementation_contract_type
            return contract_type_to_cache

        elif proxy_contract_type is not None:
            # In this case, the implementation ContactType was not discovered.
            # However, we were able to discover the ContractType of the proxy.
            # Proceed with caching the proxy; the user can update the type later
            # when the implementation is discoverable.
            self.contract_types[address] = proxy_contract_type
            return proxy_contract_type

        logger.warning(f"Unable to determine the ContractType for the proxy at '{address}'.")
        return None

    def _get_contract_type(
        self,
        address: AddressType,
        fetch_from_explorer: bool = True,
        default: Optional[ContractType] = None,
    ) -> Optional[ContractType]:
        """
        Get the _exact_ ContractType for a given address. For proxy contracts, returns
        the proxy ABIs if there are any and not the implementation ABIs.
        """
        if contract_type := self.contract_types[address]:
            return contract_type

        elif fetch_from_explorer:
            return self._get_contract_type_from_explorer(address)

        return default

    @classmethod
    def get_container(cls, contract_type: ContractType) -> ContractContainer:
        """
        Get a contract container for the given contract type.

        Args:
            contract_type (ContractType): The contract type to wrap.

        Returns:
            ContractContainer: A container object you can deploy.
        """

        return ContractContainer(contract_type)

    def instance_at(
        self,
        address: Union[str, AddressType],
        contract_type: Optional[ContractType] = None,
        txn_hash: Optional[Union[str, "HexBytes"]] = None,
        abi: Optional[Union[list[ABI], dict, str, Path]] = None,
        fetch_from_explorer: bool = True,
        proxy_info: Optional[ProxyInfoAPI] = None,
        detect_proxy: bool = True,
    ) -> ContractInstance:
        """
        Get a contract at the given address. If the contract type of the contract is known,
        either from a local deploy or a :class:`~ape.api.explorers.ExplorerAPI`, it will use that
        contract type. You can also provide the contract type from which it will cache and use
        next time.

        Raises:
            TypeError: When passing an invalid type for the `contract_type` arguments
              (expects `ContractType`).
            :class:`~ape.exceptions.ContractNotFoundError`: When the contract type is not found.

        Args:
            address (Union[str, AddressType]): The address of the plugin. If you are using the ENS
              plugin, you can also provide an ENS domain name.
            contract_type (Optional[``ContractType``]): Optionally provide the contract type
              in case it is not already known.
            txn_hash (Optional[Union[str, HexBytes]]): The hash of the transaction responsible for
              deploying the contract, if known. Useful for publishing. Defaults to ``None``.
            abi (Optional[Union[list[ABI], dict, str, Path]]): Use an ABI str, dict, path,
              or ethpm models to create a contract instance class.
            fetch_from_explorer (bool): Set to ``False`` to avoid fetching from the explorer.
              Defaults to ``True``. Won't fetch unless it needs to (uses disk & memory caching
              first).
            proxy_info (Optional[ProxyInfoAPI]): Pass in the proxy info, if it is known, to avoid
              the potentially expensive look-up.
            detect_proxy (bool): Set to ``False`` to avoid detecting if the contract is a proxy.

        Returns:
            :class:`~ape.contracts.base.ContractInstance`
        """
        if contract_type and not isinstance(contract_type, ContractType):
            prefix = f"Expected type '{ContractType.__name__}' for argument 'contract_type'"
            try:
                suffix = f"; Given '{type(contract_type).__name__}'."
            except Exception:
                suffix = "."

            raise TypeError(f"{prefix}{suffix}")

        try:
            contract_address = self.conversion_manager.convert(address, AddressType)
        except ConversionError:
            # Attempt as str.
            raise ValueError(f"Unknown address value '{address}'.")

        try:
            # Always attempt to get an existing contract type to update caches
            contract_type = self.get(
                contract_address,
                default=contract_type,
                fetch_from_explorer=fetch_from_explorer,
                proxy_info=proxy_info,
                detect_proxy=detect_proxy,
            )
        except Exception as err:
            if contract_type or abi:
                # If a default contract type was provided, don't error and use it.
                logger.error(str(err))
            else:
                raise  # Current exception

        if abi:
            # if the ABI is a str then convert it to a JSON dictionary.
            if isinstance(abi, Path) or (
                isinstance(abi, str) and "{" not in abi and Path(abi).is_file()
            ):
                # Handle both absolute and relative paths
                abi_path = Path(abi)
                if not abi_path.is_absolute():
                    abi_path = self.local_project.path / abi

                try:
                    abi = json.loads(abi_path.read_text())
                except Exception as err:
                    if contract_type:
                        # If a default contract type was provided, don't error and use it.
                        logger.error(str(err))
                    else:
                        raise  # Current exception

            elif isinstance(abi, str):
                # JSON str
                try:
                    abi = json.loads(abi)
                except Exception as err:
                    if contract_type:
                        # If a default contract type was provided, don't error and use it.
                        logger.error(str(err))
                    else:
                        raise  # Current exception

            # If the ABI was a str, it should be a list now.
            if isinstance(abi, list):
                contract_type = ContractType(abi=abi)

                # Ensure we cache the contract-types from ABI!
                self[contract_address] = contract_type

            else:
                raise TypeError(
                    f"Invalid ABI type '{type(abi)}', expecting str, list[ABI] or a JSON file."
                )

        if not contract_type:
            raise ContractNotFoundError(
                contract_address,
                self.provider.network.explorer is not None,
                self.provider.network_choice,
            )

        if not txn_hash:
            # Check for txn_hash in deployments.
            contract_name = getattr(contract_type, "name", f"{contract_type}") or ""
            deployments = self.deployments[contract_name]
            for deployment in deployments[::-1]:
                if deployment.address == contract_address and deployment.transaction_hash:
                    txn_hash = deployment.transaction_hash
                    break

        return ContractInstance(contract_address, contract_type, txn_hash=txn_hash)

    @classmethod
    def instance_from_receipt(
        cls, receipt: "ReceiptAPI", contract_type: ContractType
    ) -> ContractInstance:
        """
        A convenience method for creating instances from receipts.

        Args:
            receipt (:class:`~ape.api.transactions.ReceiptAPI`): The receipt.
            contract_type (ContractType): The deployed contract type.

        Returns:
            :class:`~ape.contracts.base.ContractInstance`
        """
        # NOTE: Mostly just needed this method to avoid a local import.
        return ContractInstance.from_receipt(receipt, contract_type)

    def get_deployments(self, contract_container: ContractContainer) -> list[ContractInstance]:
        """
        Retrieves previous deployments of a contract container or contract type.
        Locally deployed contracts are saved for the duration of the script and read from
        ``_local_deployments_mapping``, while those deployed on a live network are written to
        disk in ``deployments_map.json``.

        Args:
            contract_container (:class:`~ape.contracts.ContractContainer`): The
              ``ContractContainer`` with deployments.

        Returns:
            list[:class:`~ape.contracts.ContractInstance`]: Returns a list of contracts that
            have been deployed.
        """
        contract_type = contract_container.contract_type
        if not (contract_name := contract_type.name or ""):
            return []

        config_deployments = self._get_config_deployments(contract_name)
        if not (deployments := [*config_deployments, *self.deployments[contract_name]]):
            return []

        instances: list[ContractInstance] = []
        for deployment in deployments:
            instance = ContractInstance(
                deployment.address, contract_type, txn_hash=deployment.transaction_hash
            )
            instances.append(instance)

        return instances

    def _get_config_deployments(self, contract_name: str) -> list[Deployment]:
        if not self.network_manager.connected:
            return []

        ecosystem_name = self.provider.network.ecosystem.name
        network_name = self.provider.network.name
        all_config_deployments = (
            self.config_manager.deployments if self.config_manager.deployments else {}
        )
        ecosystem_deployments = all_config_deployments.get(ecosystem_name, {})
        network_deployments = ecosystem_deployments.get(network_name, {})
        return [
            Deployment(address=c["address"], transaction_hash=c.get("transaction_hash"))
            for c in network_deployments
            if c["contract_type"] == contract_name
        ]

    def clear_local_caches(self):
        """
        Reset local caches to a blank state.
        """
        if self.network_manager.connected:
            for cache in (
                self.contract_types,
                self.proxy_infos,
                self.contract_creations,
                self.blueprints,
            ):
                cache.memory = {}

        self.deployments.clear_local()

    def _get_contract_type_from_explorer(self, address: AddressType) -> Optional[ContractType]:
        if not self.provider.network.explorer:
            return None

        try:
            contract_type = self.provider.network.explorer.get_contract_type(address)
        except Exception as err:
            explorer_name = self.provider.network.explorer.name
            if "rate limit" in str(err).lower():
                # Don't show any additional error message during rate limit errors,
                # if it can be helped, as it may scare users into thinking their
                # contracts are not verified.
                message = str(err)
            else:
                # Carefully word this message in a way that doesn't hint at
                # any one specific reason, such as un-verified source code,
                # which is potentially a scare for users.
                message = (
                    f"Attempted to retrieve contract type from explorer '{explorer_name}' "
                    f"from address '{address}' but encountered an exception: {err}\n"
                )

            logger.error(message)
            return None

        if contract_type:
            # Cache contract so faster look-up next time.
            if not isinstance(contract_type, ContractType):
                explorer_name = self.provider.network.explorer.name
                wrong_type = type(contract_type)
                wrong_type_str = getattr(wrong_type, "__name__", f"{wrong_type}")
                logger.warning(
                    f"Explorer '{explorer_name}' returned unexpected "
                    f"type '{wrong_type_str}' ContractType."
                )
                return None

            self.contract_types[address] = contract_type

        return contract_type


def _get_combined_contract_type(
    proxy_contract_type: ContractType,
    proxy_info: ProxyInfoAPI,
    implementation_contract_type: ContractType,
) -> ContractType:
    proxy_abis = _get_relevant_additive_abis(proxy_contract_type)

    # Include "hidden" ABIs, such as Safe's `masterCopy()`.
    if proxy_info.abi and proxy_info.abi.signature not in [
        abi.signature for abi in implementation_contract_type.abi
    ]:
        proxy_abis.append(proxy_info.abi)

    return _merge_abis(implementation_contract_type, proxy_abis)


def _get_relevant_additive_abis(contract_type: ContractType) -> list[ABI]:
    # Get ABIs you would want to add to a base contract as extra,
    # such as unique ABIs from proxies.
    return [abi for abi in contract_type.abi if abi.type in ("error", "event", "function")]


def _merge_abis(base_contract: ContractType, extra_abis: list[ABI]) -> ContractType:
    contract_type = base_contract.model_copy(deep=True)
    contract_type.abi.extend(extra_abis)
    return contract_type


def _merge_contract_types(
    base_contract_type: ContractType, additive_contract_type: ContractType
) -> ContractType:
    existing_methods = set(_get_relevant_additive_abis(base_contract_type))
    relevant_abis = _get_relevant_additive_abis(additive_contract_type)
    return _merge_abis(
        base_contract_type,
        [abi for abi in relevant_abis if abi not in existing_methods],
    )



import copy
from abc import abstractmethod
from collections.abc import Collection, Iterator, Sequence
from functools import cached_property, partial
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, Optional, Union

from eth_pydantic_types import HexBytes
from eth_utils import keccak
from evmchains import PUBLIC_CHAIN_META
from pydantic import model_validator

from ape.exceptions import (
    CustomError,
    NetworkError,
    NetworkMismatchError,
    NetworkNotFoundError,
    ProviderNotConnectedError,
    ProviderNotFoundError,
)
from ape.logging import logger
from ape.types.address import AddressType
from ape.types.gas import AutoGasLimit, GasLimit
from ape.utils.basemodel import (
    BaseInterfaceModel,
    BaseModel,
    ExtraAttributesMixin,
    ExtraModelAttributes,
    ManagerAccessMixin,
)
from ape.utils.misc import (
    DEFAULT_TRANSACTION_ACCEPTANCE_TIMEOUT,
    LOCAL_NETWORK_NAME,
    log_instead_of_fail,
    raises_not_implemented,
)
from ape.utils.rpc import RPCHeaders

from .config import PluginConfig

if TYPE_CHECKING:
    from ethpm_types import ContractType
    from ethpm_types.abi import ABIType, ConstructorABI, EventABI, MethodABI

    from ape.types.address import RawAddress
    from ape.types.events import ContractLog

    from .explorers import ExplorerAPI
    from .providers import BlockAPI, ProviderAPI, UpstreamProvider
    from .trace import TraceAPI
    from .transactions import ReceiptAPI, TransactionAPI


class ProxyInfoAPI(BaseModel):
    """
    Information about a proxy contract.
    """

    target: AddressType
    """The address of the implementation contract."""

    type_name: str = ""

    @model_validator(mode="before")
    @classmethod
    def _validate_type_name(cls, model):
        if "type_name" in model:
            return model

        elif _type := model.get("type"):
            # Attempt to figure out the type name.
            if name := getattr(_type, "name", None):
                # ProxyEnum - such as from 'ape-ethereum'.
                model["type_name"] = name
            else:
                # Not sure.
                try:
                    model["type_name"] = f"{_type}"
                except Exception:
                    pass

        return model

    @log_instead_of_fail(default="<ProxyInfoAPI>")
    def __repr__(self) -> str:
        if _type := self.type_name:
            return f"<Proxy {_type} target={self.target}>"

        return "<Proxy target={self.target}"

    @property
    def abi(self) -> Optional["MethodABI"]:
        """
        Some proxies have special ABIs which may not exist in their
        contract-types by default, such as Safe's ``masterCopy()``.
        """
        return None


class EcosystemAPI(ExtraAttributesMixin, BaseInterfaceModel):
    """
    A set of related networks, such as Ethereum.
    """

    name: str
    """
    The name of the ecosystem. This should be set the same name as the plugin.
    """

    # TODO: In 0.9, make @property that returns value from config,
    #   and use REQUEST_HEADER as plugin-defined constants.
    request_header: dict = {}
    """A shareable HTTP header for network requests."""

    fee_token_symbol: str
    """The token symbol for the currency that pays for fees, such as ETH."""

    fee_token_decimals: int = 18
    """The number of the decimals the fee token has."""

    _default_network: Optional[str] = None
    """The default network of the ecosystem, such as ``local``."""

    @model_validator(mode="after")
    @classmethod
    def _validate_ecosystem(cls, model):
        headers = RPCHeaders(**model.request_header)
        headers["User-Agent"] = f"ape-{model.name}"
        model.request_header = dict(**headers)
        return model

    @log_instead_of_fail(default="<EcosystemAPI>")
    def __repr__(self) -> str:
        return f"<{self.name}>"

    @property
    def data_folder(self) -> Path:
        """
        The path to the ecosystem's data folder,
        e.g. ``$HOME/.ape/{self.name}`` unless overridden.
        """
        return self.config_manager.DATA_FOLDER / self.name

    @property
    def custom_network(self) -> "NetworkAPI":
        """
        A :class:`~ape.api.networks.NetworkAPI` for custom networks where the
        network is either not known, unspecified, or does not have an Ape plugin.
        """
        ethereum_class = None
        for plugin_name, ecosystem_class in self.plugin_manager.ecosystems:
            if plugin_name == "ethereum":
                ethereum_class = ecosystem_class
                break

        if ethereum_class is None:
            raise NetworkError("Core Ethereum plugin missing.")

        init_kwargs = {"name": "ethereum"}
        evm_ecosystem = ethereum_class(**init_kwargs)  # type: ignore
        return NetworkAPI(
            name="custom",
            ecosystem=evm_ecosystem,
            _default_provider="node",
            _is_custom=True,
        )

    @classmethod
    @abstractmethod
    def decode_address(cls, raw_address: "RawAddress") -> AddressType:
        """
        Convert a raw address to the ecosystem's native address type.

        Args:
            raw_address (:class:`~ape.types.address.RawAddress`): The address to
              convert.

        Returns:
            :class:`~ape.types.address.AddressType`
        """

    @classmethod
    @abstractmethod
    def encode_address(cls, address: AddressType) -> "RawAddress":
        """
        Convert the ecosystem's native address type to a raw integer or str address.

        Args:
            address (:class:`~ape.types.address.AddressType`): The address to convert.

        Returns:
            :class:`~ape.types.address.RawAddress`
        """

    @raises_not_implemented
    def encode_contract_blueprint(  # type: ignore[empty-body]
        self, contract_type: "ContractType", *args, **kwargs
    ) -> "TransactionAPI":
        """
        Encode a unique type of transaction that allows contracts to be created
        from other contracts, such as
        `EIP-5202 <https://eips.ethereum.org/EIPS/eip-5202>`__
        or Starknet's ``Declare`` transaction type.

        Args:
            contract_type (``ContractType``): The type of contract to create a blueprint for.
              This is the type of contract that will get created by factory contracts.
            *args (Any): Calldata, if applicable.
            **kwargs (Any): Transaction specifications, such as ``value``.

        Returns:
            :class:`~ape.ape.transactions.TransactionAPI`
        """

    @abstractmethod
    def decode_receipt(self, data: dict) -> "ReceiptAPI":
        """
        Convert data to :class:`~ape.api.transactions.ReceiptAPI`.

        Args:
            data (Dict): A dictionary of Receipt properties.

        Returns:
            :class:`~ape.api.transactions.ReceiptAPI`
        """

    @abstractmethod
    def decode_block(self, data: dict) -> "BlockAPI":
        """
        Decode data to a :class:`~ape.api.providers.BlockAPI`.

        Args:
            data (Dict): A dictionary of data to decode.

        Returns:
            :class:`~ape.api.providers.BlockAPI`
        """

    @property
    def config(self) -> PluginConfig:
        """
        The configuration of the ecosystem. See :class:`ape.managers.config.ConfigManager`
        for more information on plugin configurations.

        Returns:
            :class:`ape.api.config.PluginConfig`
        """
        return self.config_manager.get_config(self.name)

    @property
    def networks(self) -> dict[str, "NetworkAPI"]:
        """
        A dictionary of network names mapped to their API implementation.

        Returns:
            dict[str, :class:`~ape.api.networks.NetworkAPI`]
        """
        return {
            **self._networks_from_evmchains,
            **self._networks_from_plugins,
            **self._custom_networks,
        }

    @cached_property
    def _networks_from_plugins(self) -> dict[str, "NetworkAPI"]:
        return {
            network_name: network_class(name=network_name, ecosystem=self)
            for _, (ecosystem_name, network_name, network_class) in self.plugin_manager.networks
            if ecosystem_name == self.name
        }

    @cached_property
    def _networks_from_evmchains(self) -> dict[str, "NetworkAPI"]:
        # NOTE: Purposely exclude plugins here so we also prefer plugins.
        networks = {
            network_name: create_network_type(data["chainId"], data["chainId"])(
                name=network_name, ecosystem=self
            )
            for network_name, data in PUBLIC_CHAIN_META.get(self.name, {}).items()
            if network_name not in self._networks_from_plugins
        }
        forked_networks: dict[str, ForkedNetworkAPI] = {}
        for network_name, network in networks.items():
            if network_name.endswith("-fork"):
                # Already a fork.
                continue

            fork_network_name = f"{network_name}-fork"
            if any(x == fork_network_name for x in networks):
                # The forked version of this network is already known.
                continue

            forked_networks[fork_network_name] = ForkedNetworkAPI(
                name=fork_network_name, ecosystem=self
            )

        return {**networks, **forked_networks}

    @property
    def _custom_networks(self) -> dict[str, "NetworkAPI"]:
        """
        Networks from config.
        """
        networks: dict[str, NetworkAPI] = {}
        custom_networks: list[dict] = [
            n
            for n in self.network_manager.custom_networks
            if n.get("ecosystem", self.network_manager.default_ecosystem.name) == self.name
        ]

        # Ensure forks are added automatically for custom networks.
        forked_custom_networks = []
        for net in custom_networks:
            if net["name"].endswith("-fork"):
                # Already a fork.
                continue

            fork_network_name = f"{net['name']}-fork"
            if any(x["name"] == fork_network_name for x in custom_networks):
                # The forked version of this network is already known.
                continue

            # Create a forked network mirroring the custom network.
            forked_net = copy.deepcopy(net)
            forked_net["name"] = fork_network_name
            forked_custom_networks.append(forked_net)

        # NOTE: Forked custom networks are still custom networks.
        custom_networks.extend(forked_custom_networks)

        for custom_net in custom_networks:
            model_data = copy.deepcopy(custom_net)
            net_name = custom_net["name"]
            if net_name in networks:
                raise NetworkError(
                    f"More than one network named '{net_name}' in ecosystem '{self.name}'."
                )

            is_fork = net_name.endswith("-fork")
            model_data["ecosystem"] = self
            network_type = create_network_type(
                custom_net["chain_id"], custom_net["chain_id"], is_fork=is_fork
            )
            if "request_header" not in model_data:
                model_data["request_header"] = self.request_header

            network_api = network_type.model_validate(model_data)
            network_api._default_provider = custom_net.get("default_provider", "node")
            network_api._is_custom = True
            networks[net_name] = network_api

        return networks

    def __post_init__(self):
        if len(self.networks) == 0:
            raise NetworkError("Must define at least one network in ecosystem")

    def __ape_extra_attributes__(self) -> Iterator[ExtraModelAttributes]:
        yield ExtraModelAttributes(
            name="networks",
            attributes=lambda: self.networks,
            include_getattr=True,
            include_getitem=True,
        )

    def add_network(self, network_name: str, network: "NetworkAPI"):
        """
        Attach a new network to an ecosystem (e.g. L2 networks like Optimism).

        Raises:
            :class:`~ape.exceptions.NetworkError`: When the network already exists.

        Args:
            network_name (str): The name of the network to add.
            network (:class:`~ape.api.networks.NetworkAPI`): The network to add.
        """
        if network_name in self.networks:
            raise NetworkError(f"Unable to overwrite existing network '{network_name}'.")
        else:
            self.networks[network_name] = network

    @property
    def default_network_name(self) -> str:
        """
        The name of the default network in this ecosystem.

        Returns:
            str
        """
        if network := self._default_network:
            # Was set programmatically.
            return network

        networks = self.networks
        if network := self.config.get("default_network"):
            # Default found in config. Ensure is an installed network.
            if network in networks:
                return network

        if LOCAL_NETWORK_NAME in networks:
            # Default to the LOCAL_NETWORK_NAME, at last resort.
            return LOCAL_NETWORK_NAME

        elif len(networks) >= 1:
            # Use the first network.
            key = next(iter(networks.keys()))
            return networks[key].name

        # Very unlikely scenario.
        raise NetworkError("No networks found.")

    @property
    def default_network(self) -> "NetworkAPI":
        return self.get_network(self.default_network_name)

    def set_default_network(self, network_name: str):
        """
        Change the default network.

        Raises:
            :class:`~ape.exceptions.NetworkError`: When the network does not exist.

        Args:
            network_name (str): The name of the default network to switch to.
        """
        if network_name in self.networks:
            self._default_network = network_name
        else:
            raise NetworkNotFoundError(network_name, ecosystem=self.name, options=self.networks)

    @abstractmethod
    def encode_deployment(
        self, deployment_bytecode: HexBytes, abi: "ConstructorABI", *args, **kwargs
    ) -> "TransactionAPI":
        """
        Create a deployment transaction in the given ecosystem.
        This may require connecting to other networks.

        Args:
            deployment_bytecode (HexBytes): The bytecode to deploy.
            abi (ConstructorABI): The constructor interface of the contract.
            *args (Any): Constructor arguments.
            **kwargs (Any): Transaction arguments.

        Returns:
            class:`~ape.api.transactions.TransactionAPI`
        """

    @abstractmethod
    def encode_transaction(
        self, address: AddressType, abi: "MethodABI", *args, **kwargs
    ) -> "TransactionAPI":
        """
        Encode a transaction object from a contract function's ABI and call arguments.
        Additionally, update the transaction arguments with the overrides in ``kwargs``.

        Args:
            address (:class:`~ape.types.address.AddressType`): The address of the contract.
            abi (``MethodABI``): The function to call on the contract.
            *args (Any): Function arguments.
            **kwargs (Any): Transaction arguments.

        Returns:
            class:`~ape.api.transactions.TransactionAPI`
        """

    @abstractmethod
    def decode_logs(self, logs: Sequence[dict], *events: "EventABI") -> Iterator["ContractLog"]:
        """
        Decode any contract logs that match the given event ABI from the raw log data.

        Args:
            logs (Sequence[dict]): A list of raw log data from the chain.
            *events (EventABI): Event definitions to decode.

        Returns:
            Iterator[:class:`~ape.types.ContractLog`]
        """

    @raises_not_implemented
    def decode_primitive_value(  # type: ignore[empty-body]
        self, value: Any, output_type: Union[str, tuple, list]
    ) -> Union[str, HexBytes, tuple]:
        """
        Decode a primitive value-type given its ABI type as a ``str``
        and the value itself. This method is a hook for converting
        addresses, HexBytes, or other primitive data-types into
        friendlier Python equivalents.

        Args:
            value (Any): The value to decode.
            output_type (Union[str, tuple, list]): The value type.

        Returns:
            Union[str, HexBytes, tuple]
        """

    @abstractmethod
    def create_transaction(self, **kwargs) -> "TransactionAPI":
        """
        Create a transaction using key-value arguments.

        Args:
            **kwargs: Everything the transaction needs initialize.

        Returns:
            class:`~ape.api.transactions.TransactionAPI`
        """

    @abstractmethod
    def decode_calldata(self, abi: Union["ConstructorABI", "MethodABI"], calldata: bytes) -> dict:
        """
        Decode method calldata.

        Args:
            abi (Union[ConstructorABI, MethodABI]): The method called.
            calldata (bytes): The raw calldata bytes.

        Returns:
            Dict: A mapping of input names to decoded values.
            If an input is anonymous, it should use the stringified
            index of the input as the key.
        """

    @abstractmethod
    def encode_calldata(self, abi: Union["ConstructorABI", "MethodABI"], *args: Any) -> HexBytes:
        """
        Encode method calldata.

        Args:
            abi (Union[ConstructorABI, MethodABI]): The ABI of the method called.
            *args (Any): The arguments given to the method.

        Returns:
            HexBytes: The encoded calldata of the arguments to the given method.
        """

    @abstractmethod
    def decode_returndata(self, abi: "MethodABI", raw_data: bytes) -> Any:
        """
        Get the result of a contract call.

        Arg:
            abi (MethodABI): The method called.
            raw_data (bytes): Raw returned data.

        Returns:
            Any: All of the values returned from the contract function.
        """

    @raises_not_implemented
    def get_deployment_address(  # type: ignore[empty-body]
        self,
        address: AddressType,
        nonce: int,
    ) -> AddressType:
        """
        Calculate the deployment address of a contract before it is deployed.
        This is useful if the address is an argument to another contract's deployment
        and you have not yet deployed the first contract yet.
        """

    def get_network(self, network_name: str) -> "NetworkAPI":
        """
        Get the network for the given name.

        Args:
          network_name (str): The name of the network to get.

        Raises:
          :class:`~ape.exceptions.NetworkNotFoundError`: When the network is not present.

        Returns:
          :class:`~ape.api.networks.NetworkAPI`
        """
        names = {network_name, network_name.replace("-", "_"), network_name.replace("_", "-")}
        networks = self.networks
        for name in names:
            if name in networks:
                return networks[name]

            elif name == "custom":
                # Is an adhoc-custom network NOT from config.
                return self.custom_network

        raise NetworkNotFoundError(network_name, ecosystem=self.name, options=networks)

    def get_network_data(
        self, network_name: str, provider_filter: Optional[Collection[str]] = None
    ) -> dict:
        """
        Get a dictionary of data about providers in the network.

        **NOTE**: The keys are added in an opinionated order for nicely
        translating into ``yaml``.

        Args:
            network_name (str): The name of the network to get provider data from.
            provider_filter (Optional[Collection[str]]): Optionally filter the providers
              by name.

        Returns:
            dict: A dictionary containing the providers in a network.
        """
        data: dict[str, Any] = {"name": str(network_name)}

        # Only add isDefault key when True
        if network_name == self.default_network_name:
            data["isDefault"] = True

        data["providers"] = []
        network = self[network_name]

        if network.explorer:
            data["explorer"] = str(network.explorer.name)

        for provider_name in network.providers:
            if provider_filter and provider_name not in provider_filter:
                continue

            provider_data: dict = {"name": str(provider_name)}

            # Only add isDefault key when True
            if provider_name == network.default_provider_name:
                provider_data["isDefault"] = True

            data["providers"].append(provider_data)

        return data

    def get_proxy_info(self, address: AddressType) -> Optional[ProxyInfoAPI]:
        """
        Information about a proxy contract such as proxy type and implementation address.

        Args:
            address (:class:`~ape.types.address.AddressType`): The address of the contract.

        Returns:
            Optional[:class:`~ape.api.networks.ProxyInfoAPI`]: Returns ``None`` if the contract
            does not use any known proxy pattern.
        """
        return None

    def get_method_selector(self, abi: "MethodABI") -> HexBytes:
        """
        Get a contract method selector, typically via hashing such as ``keccak``.
        Defaults to using ``keccak`` but can be overridden in different ecosystems.

        Override example::

            from ape.api import EcosystemAPI
            from eth_pydantic_types import HexBytes


            class MyEcosystem(EcosystemAPI):
                def get_method_selector(self, abi: MethodABI) -> HexBytes:
                    return HexBytes(abi.selector.encode())  # Simple bytes selector

        Args:
            abi (MethodABI): The ABI object to use when calculating the
              selector bytes.

        Returns:
            HexBytes: The hashed method selector value.
        """

        return HexBytes(keccak(text=abi.selector)[:4])

    def enrich_trace(self, trace: "TraceAPI", **kwargs) -> "TraceAPI":
        """
        Enhance the data in the call tree using information about the ecosystem.

        Args:
            trace (:class:`~ape.api.trace.TraceAPI`): The trace to enrich.
            **kwargs: Additional kwargs to control enrichment, defined at the
              plugin level.

        Returns:
            :class:`~ape.api.trace.TraceAPI`
        """
        return trace

    @raises_not_implemented
    def get_python_types(  # type: ignore[empty-body]
        self, abi_type: "ABIType"
    ) -> Union[type, Sequence]:
        """
        Get the Python types for a given ABI type.

        Args:
            abi_type (``ABIType``): The ABI type to get the Python types for.

        Returns:
            Union[Type, Sequence]: The Python types for the given ABI type.
        """

    @raises_not_implemented
    def decode_custom_error(
        self,
        data: HexBytes,
        address: AddressType,
        **kwargs,
    ) -> Optional[CustomError]:
        """
        Decode a custom error class from an ABI defined in a contract.

        Args:
            data (HexBytes): The error data containing the selector
              and input data.
            address (AddressType): The address of the contract containing
              the error.
            **kwargs: Additional init kwargs for the custom error class.

        Returns:
            Optional[CustomError]: If it able to decode one, else ``None``.
        """

    def _get_request_headers(self) -> RPCHeaders:
        # Internal helper method called by NetworkManager
        headers = RPCHeaders(**self.request_header)
        # Have to do it this way to avoid "multiple-keys" error.
        configured_headers: dict = self.config.get("request_headers", {})
        for key, value in configured_headers.items():
            headers[key] = value

        return headers


class ProviderContextManager(ManagerAccessMixin):
    """
    A context manager for temporarily connecting to a network.
    When entering the context, calls the :meth:`ape.api.providers.ProviderAPI.connect` method.
    And conversely, when exiting, calls the :meth:`ape.api.providers.ProviderPAI.disconnect`
    method, unless in a multi-chain context, in which case it disconnects all providers at
    the very end of the Python session.

    The method :meth:`ape.api.networks.NetworkAPI.use_provider` returns
    an instance of this context manager.

    Usage example::

        from ape import networks

        mainnet = networks.ethereum.mainnet  # An instance of NetworkAPI
        with mainnet.use_provider("infura"):
            ...

        # Or, using choice-strings:

        with networks.parse_network_choice("ethereum:local:test"):
            ...
    """

    connected_providers: dict[str, "ProviderAPI"] = {}
    provider_stack: list[str] = []
    disconnect_map: dict[str, bool] = {}

    # We store a provider object at the class level for use when disconnecting
    # due to an exception, when interactive mode is set. If we don't hold on
    # to a reference to this object, the provider is dropped and reconnecting results
    # in losing state when using a spawned local provider
    _recycled_provider: ClassVar[Optional["ProviderAPI"]] = None

    def __init__(
        self,
        provider: "ProviderAPI",
        disconnect_after: bool = False,
        disconnect_on_exit: bool = True,
    ):
        self._provider = provider
        self._disconnect_after = disconnect_after
        self._disconnect_on_exit = disconnect_on_exit
        self._skipped_disconnect = False

    @property
    def empty(self) -> bool:
        """
        ``True`` when there are no providers in the context.
        """
        return not self.connected_providers or not self.provider_stack

    def __enter__(self, *args, **kwargs):
        # If we have a recycled provider available, this means our last exit
        # was due to an exception during interactive mode. We should resume that
        # same connection, but also clear the object so we don't do this again
        # in later provider contexts, which we would want to behave normally
        if self._recycled_provider is not None:
            # set inner var to the recycled provider for use in push_provider()
            self._provider = self._recycled_provider
            ProviderContextManager._recycled_provider = None

        return self.push_provider()

    def __exit__(self, exception, *args, **kwargs):
        if not self._disconnect_on_exit and exception is not None:
            # We want to skip disconnection when exiting due to an exception in interactive mode
            if provider := self.network_manager.active_provider:
                ProviderContextManager._recycled_provider = provider
        else:
            self.pop_provider()

    def push_provider(self):
        must_connect = not self._provider.is_connected
        if must_connect:
            self._provider.connect()

        connection_id = self._provider.connection_id
        if connection_id is None:
            raise ProviderNotConnectedError()

        self.provider_stack.append(connection_id)
        self.disconnect_map[connection_id] = self._disconnect_after
        if connection_id in self.connected_providers:
            # Using already connected instance
            if must_connect:
                # Disconnect if had to connect to check chain ID
                self._provider.disconnect()

            self._provider = self.connected_providers[connection_id]
        else:
            # Adding provider for the first time. Retain connection.
            self.connected_providers[connection_id] = self._provider

        self.network_manager.active_provider = self._provider
        return self._provider

    def pop_provider(self):
        if self.empty:
            return

        # Clear last provider
        current_id = self.provider_stack.pop()

        # Disconnect the provider in same cases.
        if self.disconnect_map.get(current_id):
            if provider := self.network_manager.active_provider:
                provider.disconnect()

            del self.disconnect_map[current_id]
            if current_id in self.connected_providers:
                del self.connected_providers[current_id]

        if not self.provider_stack:
            self.network_manager.active_provider = None
            return

        # Reset the original active provider
        prior_id = self.provider_stack[-1]
        if prior_id == current_id:
            # Active provider is not changing
            return

        if previous_provider := self.connected_providers[prior_id]:
            self.network_manager.active_provider = previous_provider

    def disconnect_all(self):
        if self.empty:
            return

        for provider in self.connected_providers.values():
            provider.disconnect()

        self.network_manager.active_provider = None
        self.connected_providers = {}


def _connect_provider(provider: "ProviderAPI") -> "ProviderAPI":
    connection_id = provider.connection_id
    if connection_id in ProviderContextManager.connected_providers:
        # Likely multi-chain testing or utilizing multiple on-going connections.
        provider = ProviderContextManager.connected_providers[connection_id]
        if not provider.is_connected:
            provider.connect()

    return provider


class NetworkAPI(BaseInterfaceModel):
    """
    A wrapper around a provider for a specific ecosystem.
    """

    name: str  # Name given when registered in ecosystem
    """The name of the network."""

    ecosystem: EcosystemAPI
    """The ecosystem of the network."""

    # TODO: In 0.9, make @property that returns value from config,
    #   and use REQUEST_HEADER as plugin-defined constants.
    request_header: dict = {}
    """A shareable network HTTP header."""

    # See ``.default_provider`` which is the proper field.
    _default_provider: str = ""

    _is_custom: bool = False

    def __repr__(self) -> str:
        try:
            chain_id = self.chain_id
        except ProviderNotConnectedError:
            # Only happens on local networks
            chain_id = None

        try:
            content = (
                f"{self.choice} chain_id={self.chain_id}" if chain_id is not None else self.choice
            )
            return f"<{content}>"
        except Exception:
            # Don't allow repr to fail.
            try:
                name = self.name
            except Exception:
                name = None

            return f"<{name}>" if name else f"{type(self)}"

    @property
    def data_folder(self) -> Path:
        """
        The path to the network's data folder,
        e.g. ``$HOME/.ape/{self.ecosystem_name}/{self.name}`` unless
        overridden.
        """
        return self.ecosystem.data_folder / self.name

    @property
    def ecosystem_config(self) -> PluginConfig:
        """
        The configuration of the network. See :class:`~ape.managers.config.ConfigManager`
        for more information on plugin configurations.
        """
        return self.ecosystem.config

    @property
    def config(self) -> PluginConfig:
        name_options = {self.name, self.name.replace("-", "_"), self.name.replace("_", "-")}
        cfg: Any
        for opt in name_options:
            if cfg := self.ecosystem_config.get(opt):
                if isinstance(cfg, dict):
                    return cfg
                elif isinstance(cfg, PluginConfig):
                    return cfg
                else:
                    raise TypeError(f"Network config must be a dictionary. Received '{type(cfg)}'.")

        return PluginConfig()

    @cached_property
    def gas_limit(self) -> GasLimit:
        return self.config.get("gas_limit", "auto")

    @cached_property
    def auto_gas_multiplier(self) -> float:
        """
        The value to multiply estimated gas by for tx-insurance.
        """
        return self.gas_limit.multiplier if isinstance(self.gas_limit, AutoGasLimit) else 1.0

    @property
    def base_fee_multiplier(self) -> float:
        """
        A multiplier to apply to a transaction base fee.
        """
        return self.config.get("base_fee_multiplier", 1.0)

    @property
    def chain_id(self) -> int:
        """
        The ID of the blockchain.

        **NOTE**: Unless overridden, returns same as
        :py:attr:`ape.api.providers.ProviderAPI.chain_id`.
        """
        if (
            self.provider.network.name == self.name
            and self.provider.network.ecosystem.name == self.ecosystem.name
        ):
            # Ensure 'active_provider' is actually (seemingly) connected
            # to this network.
            return self.provider.chain_id

        raise NetworkError(
            "Unable to reference provider to get `chain_id`: "
            f"Network '{self.name}' is detached and information is missing."
        )

    @property
    def network_id(self) -> int:
        """
        The ID of the network.

        **NOTE**: Unless overridden, returns same as
        :py:attr:`~ape.api.networks.NetworkAPI.chain_id`.
        """
        return self.chain_id

    @property
    def required_confirmations(self) -> int:
        """
        The default amount of confirmations recommended to wait
        before considering a transaction "confirmed". Confirmations
        refer to the number of blocks that have been added since the
        transaction's block.
        """
        return self.config.get("required_confirmations", 0)

    @property
    def block_time(self) -> int:
        """
        The approximate amount of time it takes for a new block to get mined to the chain.
        Configure in your ``ape-config.yaml`` file.

        Config example::

            ethereum:
              mainnet:
                block_time: 15
        """
        return self.config.get("block_time", 0)

    @property
    def transaction_acceptance_timeout(self) -> int:
        """
        The amount of time to wait for a transaction to be accepted on the network.
        Does not include waiting for block-confirmations. Defaults to two minutes.
        Local networks use smaller timeouts.
        """
        return self.config.get(
            "transaction_acceptance_timeout", DEFAULT_TRANSACTION_ACCEPTANCE_TIMEOUT
        )

    @cached_property
    def explorer(self) -> Optional["ExplorerAPI"]:
        """
        The block-explorer for the given network.

        Returns:
            :class:`ape.api.explorers.ExplorerAPI`, optional
        """
        chain_id = None if self.network_manager.active_provider is None else self.provider.chain_id
        for plugin_name, plugin_tuple in self._plugin_explorers:
            ecosystem_name, network_name, explorer_class = plugin_tuple

            # Check for explicitly configured custom networks
            plugin_config = self.config_manager.get_config(plugin_name)
            has_explorer_config = (
                plugin_config
                and self.ecosystem.name in plugin_config
                and self.name in plugin_config[self.ecosystem.name]
            )

            # Return the first registered explorer (skipping any others)
            if self.ecosystem.name == ecosystem_name and (
                self.name == network_name or has_explorer_config
            ):
                return explorer_class(name=plugin_name, network=self)

            elif chain_id is not None and explorer_class.supports_chain(chain_id):
                # NOTE: Adhoc networks will likely reach here.
                return explorer_class(name=plugin_name, network=self)

        return None  # May not have an block explorer

    @property
    def _plugin_explorers(self) -> list[tuple]:
        # Abstracted for testing purposes.
        return self.plugin_manager.explorers

    @property
    def is_mainnet(self) -> bool:
        """
        True when the network is the mainnet network for the ecosystem.
        """
        cfg_is_mainnet: Optional[bool] = self.config.get("is_mainnet")
        if cfg_is_mainnet is not None:
            return cfg_is_mainnet

        return self.name == "mainnet"

    @property
    def is_fork(self) -> bool:
        """
        True when using a forked network.
        """
        return self.name.endswith("-fork")

    @property
    def is_local(self) -> bool:
        """
        True when using the local network.
        """
        return self.name == LOCAL_NETWORK_NAME

    @property
    def is_dev(self) -> bool:
        """
        True when using a local network, including forks.
        """
        return self.is_local or self.is_fork

    @property
    def is_adhoc(self) -> bool:
        """
        Is a custom network from CLI only, e.g. was not configured
        in any CLI value and is mostly an "unknown" network.
        """
        return self.name == "custom" and not self._is_custom

    @property
    def is_custom(self) -> bool:
        """
        True when this network is a configured custom network.
        """
        return self._is_custom

    @cached_property
    def providers(self):  # -> dict[str, Partial[ProviderAPI]]
        """
        The providers of the network, such as Infura, Alchemy, or Node.

        Returns:
            dict[str, partial[:class:`~ape.api.providers.ProviderAPI`]]
        """
        providers = {}
        for _, plugin_tuple in self._get_plugin_providers():
            ecosystem_name, network_name, provider_class = plugin_tuple
            provider_name = (
                provider_class.__module__.split(".")[0].replace("_", "-").replace("ape-", "")
            )

            is_custom_with_config = self._is_custom and self.default_provider_name == provider_name
            # NOTE: Custom networks that are NOT from config must work with any provider.
            #    Also, ensure we are only adding forked providers for forked networks and
            #    non-forking providers for non-forked networks. For custom networks, it
            #    can be trickier (see last condition).
            # TODO: In 0.9, add a better way for class-level ForkedProviders to define
            #   themselves as "Fork" providers.
            if (
                (self.ecosystem.name == ecosystem_name and self.name == network_name)
                or self.is_adhoc
                or (
                    is_custom_with_config
                    and (
                        (self.is_fork and "Fork" in provider_class.__name__)
                        or (not self.is_fork and "Fork" not in provider_class.__name__)
                    )
                    and provider_class.__name__
                    == "Node"  # Ensure uses Node class instead of GethDev
                )
            ):
                # NOTE: Lazily load provider config
                provider_name = provider_class.NAME or provider_name
                providers[provider_name] = partial(
                    provider_class,
                    name=provider_name,
                    network=self,
                )

        # Any EVM-chain works with node provider.
        if "node" not in providers and self.name in self.ecosystem._networks_from_evmchains:
            # NOTE: Arbitrarily using sepolia to access the Node class.
            node_provider_cls = self.network_manager.ethereum.sepolia.get_provider("node").__class__
            providers["node"] = partial(node_provider_cls, name="node", network=self)

        return providers

    def _get_plugin_providers(self):
        # NOTE: Abstracted for testing purposes.
        return self.plugin_manager.providers

    def _get_plugin_provider_names(self) -> Iterator[str]:
        for _, plugin_tuple in self._get_plugin_providers():
            ecosystem_name, network_name, provider_class = plugin_tuple
            yield provider_class.__module__.split(".")[0].replace("_", "-").replace("ape-", "")

    def get_provider(
        self,
        provider_name: Optional[str] = None,
        provider_settings: Optional[dict] = None,
        connect: bool = False,
    ):
        """
        Get a provider for the given name. If given ``None``, returns the default provider.

        Args:
            provider_name (str, optional): The name of the provider to get. Defaults to ``None``.
              When ``None``, returns the default provider.
            provider_settings (dict, optional): Settings to apply to the provider. Defaults to
              ``None``.
            connect (bool): Set to ``True`` when you also want the provider to connect.

        Returns:
            :class:`~ape.api.providers.ProviderAPI`
        """
        if not (provider_name := provider_name or self.default_provider_name):
            raise NetworkError(
                f"No default provider for network '{self.name}'. "
                "Set one in your pyproject.toml/ape-config.yaml file:\n"
                f"\n{self.ecosystem.name}:"
                f"\n  {self.name}:"
                "\n    default_provider: <DEFAULT_PROVIDER>"
            )
        provider_settings = provider_settings or {}
        if ":" in provider_name:
            # NOTE: Shortcut that allows `--network ecosystem:network:http://...` to work
            provider_settings["uri"] = provider_name
            provider_name = "node"

        elif provider_name.endswith(".ipc"):
            provider_settings["ipc_path"] = provider_name
            provider_name = "node"

        if provider_name in self.providers:
            provider = self.providers[provider_name](provider_settings=provider_settings)
            return _connect_provider(provider) if connect else provider

        elif self.is_fork:
            # If it can fork Ethereum (and we are asking for it) assume it can fork this one.
            # TODO: Refactor this approach to work for custom-forked non-EVM networks.
            common_forking_providers = self.network_manager.ethereum.mainnet_fork.providers
            if provider_name in common_forking_providers:
                provider = common_forking_providers[provider_name](
                    provider_settings=provider_settings,
                    network=self,
                )
                return _connect_provider(provider) if connect else provider

        raise ProviderNotFoundError(
            provider_name,
            network=self.name,
            ecosystem=self.ecosystem.name,
            options=self.providers,
        )

    def use_provider(
        self,
        provider: Union[str, "ProviderAPI"],
        provider_settings: Optional[dict] = None,
        disconnect_after: bool = False,
        disconnect_on_exit: bool = True,
    ) -> ProviderContextManager:
        """
        Use and connect to a provider in a temporary context. When entering the context, it calls
        method :meth:`ape.api.providers.ProviderAPI.connect` and when exiting, it calls
        method :meth:`ape.api.providers.ProviderAPI.disconnect`.

        Usage example::

            from ape import networks

            mainnet = networks.ethereum.mainnet  # An instance of NetworkAPI
            with mainnet.use_provider("infura"):
                ...

        Args:
            provider (Union[str, :class:`~ape.api.providers.ProviderAPI`]): The provider
              instance or the name of the provider to use.
            provider_settings (dict, optional): Settings to apply to the provider.
              Defaults to ``None``.
            disconnect_after (bool): Set to ``True`` to force a disconnect after ending
              the context. This defaults to ``False`` so you can re-connect to the
              same network, such as in a multi-chain testing scenario.
            disconnect_on_exit (bool): Whether to disconnect on the exit of the python
              session. Defaults to ``True``.

        Returns:
            :class:`~ape.api.networks.ProviderContextManager`
        """

        settings = provider_settings or {}

        # NOTE: The main reason we allow a provider instance here is to avoid unnecessarily
        #   re-initializing the class.
        provider_obj = (
            self.get_provider(provider_name=provider, provider_settings=settings, connect=True)
            if isinstance(provider, str)
            else provider
        )

        return ProviderContextManager(
            provider=provider_obj,
            disconnect_after=disconnect_after,
            disconnect_on_exit=disconnect_on_exit,
        )

    @property
    def default_provider_name(self) -> Optional[str]:
        """
        The name of the default provider or ``None``.

        Returns:
            Optional[str]
        """

        provider_from_config: str
        if provider := self._default_provider:
            # Was set programmatically.
            return provider

        elif provider_from_config := self.config.get("default_provider"):
            # The default is found in the Network's config class.
            return provider_from_config

        elif len(self.providers) > 0:
            # No default set anywhere - use the first installed.
            return list(self.providers)[0]

        # There are no providers at all for this network.
        return None

    @property
    def default_provider(self) -> Optional["ProviderAPI"]:
        if (name := self.default_provider_name) and name in self.providers:
            return self.get_provider(name)

        return None

    @property
    def choice(self) -> str:
        return f"{self.ecosystem.name}:{self.name}"

    def set_default_provider(self, provider_name: str):
        """
        Change the default provider.

        Raises:
            :class:`~ape.exceptions.NetworkError`: When the given provider is not found.

        Args:
            provider_name (str): The name of the provider to switch to.
        """

        if provider_name in self.providers:
            self._default_provider = provider_name
        else:
            raise NetworkError(f"Provider '{provider_name}' not found in network '{self.choice}'.")

    def use_default_provider(
        self,
        provider_settings: Optional[dict] = None,
        disconnect_after: bool = False,
    ) -> ProviderContextManager:
        """
        Temporarily connect and use the default provider. When entering the context, it calls
        method :meth:`ape.api.providers.ProviderAPI.connect` and when exiting, it calls
        method :meth:`ape.api.providers.ProviderAPI.disconnect`.

        **NOTE**: If multiple providers exist, uses whatever was "first" registered.

        Usage example::

            from ape import networks

            mainnet = networks.ethereum.mainnet  # An instance of NetworkAPI
            with mainnet.use_default_provider():
                ...

        Args:
            provider_settings (dict, optional): Settings to override the provider.
            disconnect_after (bool): Set to ``True`` to force a disconnect after ending
              the context. This defaults to ``False`` so you can re-connect to the
              same network, such as in a multi-chain testing scenario.

        Returns:
            :class:`~ape.api.networks.ProviderContextManager`
        """
        if self.default_provider:
            settings = provider_settings or {}
            return self.use_provider(
                self.default_provider.name,
                provider_settings=settings,
                disconnect_after=disconnect_after,
            )

        raise NetworkError(f"No providers for network '{self.name}'.")

    def publish_contract(self, address: AddressType):
        """
        A convenience method to publish a contract to the explorer.

        Raises:
            :class:`~ape.exceptions.NetworkError`: When there is no explorer for this network.

        Args:
            address (:class:`~ape.types.address.AddressType`): The address of the contract.
        """
        if not self.explorer:
            raise NetworkError("Unable to publish contract - no explorer plugin installed.")

        logger.info(f"Publishing and verifying contract using '{self.explorer.name}'.")
        self.explorer.publish_contract(address)

    def verify_chain_id(self, chain_id: int):
        """
        Verify a chain ID for this network.

        Args:
            chain_id (int): The chain ID to verify.

        Raises:
            :class:`~ape.exceptions.NetworkMismatchError`: When the network is
              not local or adhoc and has a different hardcoded chain ID than
              the given one.
        """
        if self.name not in ("custom", LOCAL_NETWORK_NAME) and self.chain_id != chain_id:
            raise NetworkMismatchError(chain_id, self)

    def _get_request_headers(self) -> RPCHeaders:
        # Internal helper method called by NetworkManager
        headers = RPCHeaders(**self.request_header)
        # Have to do it this way to avoid multiple-keys error.
        configured_headers: dict = self.config.get("request_headers", {})
        for key, value in configured_headers.items():
            headers[key] = value

        return headers


class ForkedNetworkAPI(NetworkAPI):
    @property
    def upstream_network(self) -> NetworkAPI:
        """
        The network being forked.
        """
        network_name = self.name.replace("-fork", "").replace("_fork", "")
        return self.ecosystem.get_network(network_name)

    @property
    def upstream_provider(self) -> "UpstreamProvider":
        """
        The provider used when requesting data before the local fork.
        Set this in your config under the network settings.
        When not set, will attempt to use the default provider, if one
        exists.
        """
        config_choice: str = self.config.get("upstream_provider")
        if provider_name := config_choice or self.upstream_network.default_provider_name:
            return self.upstream_network.get_provider(provider_name)

        raise NetworkError(f"Upstream network '{self.upstream_network}' has no providers.")

    @property
    def upstream_chain_id(self) -> int:
        """
        The chain Id of the upstream network.
        For example, when on ``mainnet-fork``, this should always
        return the chain ID for ``mainnet``. Some providers may use
        a different chain ID for forked networks while some do not.
        This property should ALWAYS be that of the forked network, regardless.
        """
        return self.upstream_network.chain_id

    def use_upstream_provider(self) -> ProviderContextManager:
        """
        Connect to the upstream provider.

        Returns:
            :class:`~ape.api.networks.ProviderContextManager`
        """
        return self.upstream_network.use_provider(self.upstream_provider)


def create_network_type(chain_id: int, network_id: int, is_fork: bool = False) -> type[NetworkAPI]:
    """
    Easily create a :class:`~ape.api.networks.NetworkAPI` subclass.
    """
    BaseNetwork = ForkedNetworkAPI if is_fork else NetworkAPI

    class network_def(BaseNetwork):  # type: ignore
        @property
        def chain_id(self) -> int:
            return chain_id

        @property
        def network_id(self) -> int:
            return network_id

    return network_def


# TODO: Can remove in 0.9 since `LOCAL_NETWORK_NAME` doesn't need to be here.
__all__ = [
    "LOCAL_NETWORK_NAME",  # Have to leave for backwards compat.
    "EcosystemAPI",
    "ForkedNetworkAPI",
    "NetworkAPI",
    "ProviderContextManager",
    "ProxyInfoAPI",
    "create_network_type",
]



def test_append_gas(gas_tracker, owner, vyper_contract_instance, eth_tester_provider):
    tx = vyper_contract_instance.setNumber(924, sender=owner)
    trace = tx.trace
    gas_tracker.append_gas(trace, vyper_contract_instance.address)
    report = gas_tracker.session_gas_report
    contract_name = vyper_contract_instance.contract_type.name
    assert contract_name in report
    assert "setNumber" in report[contract_name]
    assert tx.gas_used in report[contract_name]["setNumber"]


def test_append_gas_deploy(gas_tracker, vyper_contract_instance):
    tx = vyper_contract_instance.creation_metadata.receipt
    trace = tx.trace
    gas_tracker.append_gas(trace, vyper_contract_instance.address)
    report = gas_tracker.session_gas_report
    contract_name = vyper_contract_instance.contract_type.name
    assert contract_name in report
    assert "__new__" in report[contract_name]
    assert tx.gas_used in report[contract_name]["__new__"]


def test_append_gas_transfer(gas_tracker, sender, receiver):
    tx = sender.transfer(receiver, 0)
    trace = tx.trace
    gas_tracker.append_gas(trace, receiver.address)
    report = gas_tracker.session_gas_report

    # ETH-transfers are not included in the final report.
    assert report is None



import pytest

from ape.utils.rpc import RPCHeaders


class TestRPCHeaders:
    @pytest.fixture
    def headers(self):
        return RPCHeaders()

    @pytest.mark.parametrize("key", ("Content-Type", "CONTENT-TYPE"))
    def test_setitem_key_case_insensitive(self, key, headers):
        headers[key] = "application/javascript"
        headers[key.lower()] = "application/json"
        assert headers[key] == "application/json"
        assert headers[key.lower()] == "application/json"

    def test_setitem_user_agent_does_not_add_twice(self, headers):
        expected = "test-user-agent/1.0"
        headers["User-Agent"] = expected
        # Add again. It should not add twice.
        headers["User-Agent"] = expected
        assert headers["User-Agent"] == expected

    def test_setitem_user_agent_appends(self, headers):
        headers["User-Agent"] = "test0/1.0"
        headers["User-Agent"] = "test1/2.0"
        assert headers["User-Agent"] == "test0/1.0 test1/2.0"

    def test_setitem_user_agent_parts_exist(self, headers):
        """
        Tests the case when user-agents share a sub-set
        of each other, that it does not duplicate.
        """
        headers["User-Agent"] = "test0/1.0"
        # The beginning of the user-agent is already present.
        # It shouldn't add the full thing.
        headers["User-Agent"] = "test0/1.0 test1/2.0"
        assert headers["User-Agent"] == "test0/1.0 test1/2.0"
        # unexpected = "test0/1.0 test0/1.0 test1/2.0"

    @pytest.mark.parametrize("key", ("user-agent", "User-Agent", "USER-AGENT"))
    def test_contains_user_agent(self, key, headers):
        headers["User-Agent"] = "test0/1.0"
        assert key in headers



from typing import TYPE_CHECKING, Optional

from eth_pydantic_types import HexBytes

from ape.contracts.base import ContractContainer
from ape_ethereum.proxies import ProxyType
from ape_test.provider import LocalProvider

if TYPE_CHECKING:
    from ape.types import AddressType, BlockID

"""
NOTE: Most proxy tests are in `geth/test_proxy.py`.
"""


def test_minimal_proxy(ethereum, minimal_proxy_container, chain, owner):
    placeholder = "0xBEbeBeBEbeBebeBeBEBEbebEBeBeBebeBeBebebe"
    if placeholder in chain.contracts:
        del chain.contracts[placeholder]

    minimal_proxy = owner.deploy(minimal_proxy_container, sender=owner)
    chain.provider.network.__dict__["explorer"] = None  # Ensure no explorer, messes up test.
    actual = ethereum.get_proxy_info(minimal_proxy.address)
    assert actual is not None
    assert actual.type == ProxyType.Minimal
    # It is the placeholder value still.
    assert actual.target == placeholder

    # Show getting the contract using the proxy address.
    contract = chain.contracts.instance_at(minimal_proxy.address)
    abi = contract.contract_type.abi
    if isinstance(abi, list):
        assert abi == []
    # else: is messed up from other test (xdist).


def test_provider_not_supports_get_storage(
    get_contract_type, owner, vyper_contract_instance, ethereum, chain, networks
):
    """
    The get storage slot RPC is required to detect this proxy, so it won't work
    on EthTester provider. However, we can make sure that it doesn't try to
    call `get_storage()` more than once.
    """

    class MyProvider(LocalProvider):
        times_get_storage_was_called: int = 0

        def get_storage(  # type: ignore[empty-body]
            self, address: "AddressType", slot: int, block_id: Optional["BlockID"] = None
        ) -> "HexBytes":
            self.times_get_storage_was_called += 1
            raise NotImplementedError()

    my_provider = MyProvider(name="test", network=ethereum.local)
    my_provider._web3 = chain.provider._web3

    _type = get_contract_type("beacon")
    beacon_contract = ContractContainer(_type)
    target = vyper_contract_instance.address
    beacon_instance = owner.deploy(beacon_contract, target)
    beacon = beacon_instance.address

    _type = get_contract_type("BeaconProxy")
    contract = ContractContainer(_type)
    contract_instance = owner.deploy(contract, beacon, HexBytes(""))

    # Ensure not already cached.
    if contract_instance.address in chain.contracts.proxy_infos:
        del chain.contracts.proxy_infos[contract_instance.address]

    init_provider = networks.active_provider
    networks.active_provider = my_provider
    try:
        actual = ethereum.get_proxy_info(contract_instance.address)
    finally:
        networks.active_provider = init_provider

    assert actual is None  # Because of provider.
    assert my_provider.times_get_storage_was_called == 1



import re
from re import Pattern
from typing import Union

from pydantic import field_serializer, field_validator
from pydantic_settings import SettingsConfigDict

from ape.api.config import ConfigEnum, PluginConfig
from ape.utils.misc import SOURCE_EXCLUDE_PATTERNS


class OutputExtras(ConfigEnum):
    """
    Extra stuff you can output. It will
    appear in ``.build/{key.lower()/``
    """

    ABI = "ABI"
    """
    Include this value to output the ABIs of your contracts
    to minified JSONs. This is useful for hosting purposes
    for web-apps.
    """


class Config(PluginConfig):
    """
    Configure general compiler settings.
    """

    exclude: set[Union[str, Pattern]] = set()
    """
    Source exclusion globs or regex patterns across all file types.
    To use regex, start your values with ``r"`` and they'll be turned
    into regex pattern objects.

    **NOTE**: ``ape.utils.misc.SOURCE_EXCLUDE_PATTERNS`` are automatically
    included in this set.
    """

    include_dependencies: bool = False
    """
    Set to ``True`` to compile dependencies during ``ape compile``.
    Generally, dependencies are not compiled during ``ape compile``
    This is because dependencies may not compile in Ape on their own,
    but you can still reference them in your project's contracts' imports.
    Some projects may be more dependency-based and wish to have the
    contract types always compiled during ``ape compile``, and these projects
    should configure ``include_dependencies`` to be ``True``.
    """

    output_extra: list[OutputExtras] = []
    """
    Extra selections to output. Outputs to ``.build/{key.lower()}``.
    """

    model_config = SettingsConfigDict(extra="allow", env_prefix="APE_COMPILE_")

    @field_validator("exclude", mode="before")
    @classmethod
    def validate_exclude(cls, value):
        given_values = []

        # Convert regex to Patterns.
        for given in value or []:
            if (given.startswith('r"') and given.endswith('"')) or (
                given.startswith("r'") and given.endswith("'")
            ):
                value_clean = given[2:-1]
                pattern = re.compile(value_clean)
                given_values.append(pattern)

            else:
                given_values.append(given)

        # Include defaults.
        return {*given_values, *SOURCE_EXCLUDE_PATTERNS}

    @field_serializer("exclude", when_used="json")
    def serialize_exclude(self, exclude, info):
        """
        Exclude is put back with the weird r-prefix so we can
        go to-and-from.
        """
        result: list[str] = []
        for excl in exclude:
            if isinstance(excl, Pattern):
                result.append(f'r"{excl.pattern}"')
            else:
                result.append(excl)

        return result



import pytest

from ape.pytest.config import ConfigWrapper
from ape.pytest.fixtures import ReceiptCapture


@pytest.fixture
def pytest_config(mocker):
    return mocker.MagicMock()


@pytest.fixture
def config_wrapper(pytest_config):
    return ConfigWrapper(pytest_config)


@pytest.fixture
def receipt_capture(config_wrapper):
    return ReceiptCapture(config_wrapper)


def test_when_txn_hash_not_exists_does_not_error(receipt_capture):
    actual = receipt_capture.capture("123")
    assert actual is None



import re
from collections.abc import Callable, Iterator, Sequence
from enum import Enum
from functools import cache, cached_property
from importlib import import_module
from typing import TYPE_CHECKING, Any, Optional, Union

import click
from click import Choice, Context, Parameter

from ape.exceptions import (
    AccountsError,
    EcosystemNotFoundError,
    NetworkNotFoundError,
    ProviderNotFoundError,
)

if TYPE_CHECKING:
    from ape.api.accounts import AccountAPI
    from ape.api.providers import ProviderAPI

_ACCOUNT_TYPE_FILTER = Union[
    None, Sequence["AccountAPI"], type["AccountAPI"], Callable[["AccountAPI"], bool]
]


def _get_accounts(key: _ACCOUNT_TYPE_FILTER) -> list["AccountAPI"]:
    from ape.utils.basemodel import ManagerAccessMixin as access

    accounts = access.account_manager

    add_test_accounts = False
    if key is None:
        account_list = list(accounts)

        # Include test accounts at end.
        add_test_accounts = True

    elif isinstance(key, type):
        # Filtering by type.
        account_list = accounts.get_accounts_by_type(key)

    elif isinstance(key, (list, tuple, set)):
        # Given an account list.
        account_list = key  # type: ignore

    else:
        # Filtering by callable.
        account_list = [a for a in accounts if key(a)]  # type: ignore

    sorted_accounts = sorted(account_list, key=lambda a: a.alias or "")
    if add_test_accounts:
        sorted_accounts.extend(accounts.test_accounts)

    return sorted_accounts


class Alias(click.Choice):
    """
    A ``click.Choice`` for loading account aliases for the active project at runtime.

    Provide an ``account_type`` to limit the type of account to choose from.
    Defaults to all account types in ``choices()``.
    """

    name = "alias"

    def __init__(self, key: _ACCOUNT_TYPE_FILTER = None):
        # NOTE: we purposely skip the constructor of `Choice`
        self.case_sensitive = False
        self._key_filter = key

    @cached_property
    def choices(self) -> Sequence:  # type: ignore[override]
        from ape.types.basic import _LazySequence

        return _LazySequence(self._choices_iterator)

    @property
    def _choices_iterator(self) -> Iterator[str]:
        for acct in _get_accounts(key=self._key_filter):
            if acct.alias is None:
                continue

            yield acct.alias


class PromptChoice(click.ParamType):
    """
    A choice option or argument from user selection.

    Usage example::

        def choice_callback(ctx, param, value):
            return param.type.get_user_selected_choice()


        @click.command()
        @click.option(
            "--choice",
            type=PromptChoice(["foo", "bar"]),
            callback=choice_callback,
        )
        def cmd(choice):
            click.echo(f"__expected_{choice}")
    """

    def __init__(self, choices: Sequence[str], name: Optional[str] = None):
        self.choices = choices
        # Since we purposely skip the super() constructor, we need to make
        # sure the class still has a name.
        self.name = name or "option"

    def print_choices(self):
        """
        Echo the choices to the terminal.
        """
        choices = dict(enumerate(self.choices, 0))
        did_print = False
        for idx, choice in choices.items():
            click.echo(f"{idx}. {choice}")
            did_print = True

        if did_print:
            click.echo()

    def convert(
        self, value: Any, param: Optional[Parameter], ctx: Optional[Context]
    ) -> Optional[Any]:
        # noinspection PyBroadException
        try:
            choice_index = int(value)
            if choice_index < 0:
                self.fail("Invalid choice", param=param)

            return self.choices[choice_index]

        except Exception:
            return self.fail_from_invalid_choice(param)

    def fail_from_invalid_choice(self, param):
        return self.fail("Invalid choice.", param=param)

    def select(self) -> str:
        choices = "\n".join(self.choices)
        choice = click.prompt(f"Select one of the following:\n{choices}").strip()
        if not choice.isnumeric():
            return choice

        # User input an index.
        choice_idx = int(choice)
        if 0 <= choice_idx < len(self.choices):
            return self.choices[choice_idx]

        raise IndexError(f"Choice index '{choice_idx}' out of range.")


def select_account(
    prompt_message: Optional[str] = None, key: _ACCOUNT_TYPE_FILTER = None
) -> "AccountAPI":
    """
    Prompt the user to pick from their accounts and return that account.
    Use this method if you want to prompt users to select accounts _outside_
    of CLI options. For CLI options, use
    :meth:`~ape.cli.options.account_option`.

    Args:
        prompt_message (Optional[str]): Customize the prompt message.
        key (Union[None, type[AccountAPI], Callable[[AccountAPI], bool]]):
          If given, the user may only select a matching account. You can provide
          a list of accounts, an account class type, or a callable for filtering
          the accounts.

    Returns:
        :class:`~ape.api.accounts.AccountAPI`
    """
    from ape.api.accounts import AccountAPI

    if key and isinstance(key, type) and not issubclass(key, AccountAPI):
        raise AccountsError(f"Cannot return accounts with type '{key}'.")

    prompt = AccountAliasPromptChoice(prompt_message=prompt_message, key=key)
    return prompt.select_account()


class AccountAliasPromptChoice(PromptChoice):
    """
    Prompts the user to select an alias from their accounts.
    Useful for adhoc scripts to lessen the need to hard-code aliases.
    """

    def __init__(
        self,
        key: _ACCOUNT_TYPE_FILTER = None,
        prompt_message: Optional[str] = None,
        name: str = "account",
    ):
        # NOTE: we purposely skip the constructor of `PromptChoice`
        self._key_filter = key
        self._prompt_message = prompt_message or "Select an account"
        self.name = name

    @cached_property
    def choices(self) -> Sequence[str]:  # type: ignore[override]
        from ape.types.basic import _LazySequence

        return _LazySequence(self._choices_iterator)

    def convert(
        self, value: Any, param: Optional[Parameter], ctx: Optional[Context]
    ) -> Optional["AccountAPI"]:
        if value is None:
            return None

        if isinstance(value, str) and value.isnumeric():
            alias = super().convert(value, param, ctx)
        else:
            alias = value

        from ape.utils.basemodel import ManagerAccessMixin as access

        accounts = access.account_manager
        if isinstance(alias, str) and alias.upper().startswith("TEST::"):
            idx_str = alias.upper().replace("TEST::", "")
            if not idx_str.isnumeric():
                if alias in accounts.aliases:
                    # Was actually a similar-alias.
                    return accounts.load(alias)

                self.fail(f"Cannot reference test account by '{value}'.", param=param)

            account_idx = int(idx_str)
            if 0 <= account_idx < len(accounts.test_accounts):
                return accounts.test_accounts[int(idx_str)]

            self.fail(f"Index '{idx_str}' is not valid.", param=param)

        elif alias and alias in accounts.aliases:
            return accounts.load(alias)

        self.fail(f"Account with alias '{alias}' not found.", param=param)

    def print_choices(self):
        choices = dict(enumerate(self.choices, 0))
        did_print = False
        for idx, choice in choices.items():
            if not choice.startswith("TEST::"):
                click.echo(f"{idx}. {choice}")
                did_print = True

        from ape.utils.basemodel import ManagerAccessMixin as access

        accounts = access.account_manager
        len_test_accounts = len(accounts.test_accounts) - 1
        if len_test_accounts > 0:
            msg = "'TEST::account_idx', where `account_idx` is in [0..{len_test_accounts}]\n"
            if did_print:
                msg = f"Or {msg}"

            click.echo(msg)

        elif did_print:
            click.echo()

    @property
    def _choices_iterator(self) -> Iterator[str]:
        # NOTE: Includes test accounts unless filtered out by key.
        for account in _get_accounts(key=self._key_filter):
            if account and (alias := account.alias):
                yield alias

    def select_account(self) -> "AccountAPI":
        """
        Returns the selected account.

        Returns:
            :class:`~ape.api.accounts.AccountAPI`
        """
        from ape.utils.basemodel import ManagerAccessMixin as access

        accounts = access.account_manager
        if not self.choices or len(self.choices) == 0:
            raise AccountsError("No accounts found.")
        elif len(self.choices) == 1 and self.choices[0].startswith("TEST::"):
            return accounts.test_accounts[int(self.choices[0].replace("TEST::", ""))]
        elif len(self.choices) == 1:
            return accounts.load(self.choices[0])

        self.print_choices()
        return click.prompt(self._prompt_message, type=self)

    def fail_from_invalid_choice(self, param):
        return self.fail("Invalid choice. Type the number or the alias.", param=param)


_NETWORK_FILTER = Optional[Union[list[str], str]]
_NONE_NETWORK = "__NONE_NETWORK__"


def get_networks(
    ecosystem: _NETWORK_FILTER = None,
    network: _NETWORK_FILTER = None,
    provider: _NETWORK_FILTER = None,
) -> Sequence:
    # NOTE: Use str-keys and lru_cache.
    return _get_networks_sequence_from_cache(
        _network_filter_to_key(ecosystem),
        _network_filter_to_key(network),
        _network_filter_to_key(provider),
    )


@cache
def _get_networks_sequence_from_cache(ecosystem_key: str, network_key: str, provider_key: str):
    networks = import_module("ape.utils.basemodel").ManagerAccessMixin.network_manager
    module = import_module("ape.types.basic")
    return module._LazySequence(
        networks.get_network_choices(
            ecosystem_filter=_key_to_network_filter(ecosystem_key),
            network_filter=_key_to_network_filter(network_key),
            provider_filter=_key_to_network_filter(provider_key),
        )
    )


def _network_filter_to_key(filter_: _NETWORK_FILTER) -> str:
    if filter_ is None:
        return "__none__"

    elif isinstance(filter_, list):
        return ",".join(filter_)

    return filter_


def _key_to_network_filter(key: str) -> _NETWORK_FILTER:
    if key == "__none__":
        return None

    elif "," in key:
        return [n.strip() for n in key.split(",")]

    return key


class NetworkChoice(click.Choice):
    """
    A ``click.Choice`` to provide network choice defaults for the active project.

    Optionally provide a list of ecosystem names, network names, or provider names
    to filter the results by.

    This is used in :meth:`~ape.cli.options.network_option`.
    """

    CUSTOM_NETWORK_PATTERN = re.compile(r"\w*:\w*:(https?|wss?)://\w*.*|.*\.ipc")

    def __init__(
        self,
        case_sensitive=True,
        ecosystem: _NETWORK_FILTER = None,
        network: _NETWORK_FILTER = None,
        provider: _NETWORK_FILTER = None,
        base_type: Optional[Union[type, str]] = None,
        callback: Optional[Callable] = None,
    ):
        self._base_type = base_type
        self.callback = callback
        self.case_sensitive = case_sensitive
        self.ecosystem = ecosystem
        self.network = network
        self.provider = provider
        # NOTE: Purposely avoid super().init for performance reasons.

    @property
    def base_type(self) -> Union[type["ProviderAPI"], str]:
        if self._base_type is not None:
            return self._base_type

        # perf: Keep base-type as a forward-ref when only using the default.
        #   so things load faster.
        self._base_type = "ProviderAPI"
        return self._base_type

    @base_type.setter
    def base_type(self, value):
        self._base_type = value

    @cached_property
    def choices(self) -> Sequence[Any]:  # type: ignore[override]
        return get_networks(ecosystem=self.ecosystem, network=self.network, provider=self.provider)

    def get_metavar(self, param):
        return "[ecosystem-name][:[network-name][:[provider-name]]]"

    def convert(self, value: Any, param: Optional[Parameter], ctx: Optional[Context]) -> Any:
        if not value or value.lower() in ("none", "null"):
            return self.callback(ctx, param, _NONE_NETWORK) if self.callback else _NONE_NETWORK

        if self.base_type == "ProviderAPI" or isinstance(self.base_type, type):
            # Return the provider.
            from ape.utils.basemodel import ManagerAccessMixin as access

            networks = access.network_manager
            try:
                value = networks.get_provider_from_choice(network_choice=value)
            except (EcosystemNotFoundError, NetworkNotFoundError, ProviderNotFoundError) as err:
                self.fail(str(err))

        return self.callback(ctx, param, value) if self.callback else value


class OutputFormat(Enum):
    """
    An enum representing output formats, such as ``TREE`` or ``YAML``.
    Use this to select a subset of common output formats to use
    when creating a :meth:`~ape.cli.choices.output_format_choice`.
    """

    TREE = "TREE"
    """A rich text tree view of the data."""

    YAML = "YAML"
    """A standard .yaml format of the data."""


def output_format_choice(options: Optional[list[OutputFormat]] = None) -> Choice:
    """
    Returns a ``click.Choice()`` type for the given options.

    Args:
        options (list[:class:`~ape.choices.OutputFormat`], optional):
          Limit the formats to accept. Defaults to allowing all formats.

    Returns:
        :class:`click.Choice`
    """

    options = options or list(OutputFormat)

    # Uses `str` form of enum for CLI choices.
    return click.Choice([o.value for o in options], case_sensitive=False)


class LazyChoice(Choice):
    """
    A simple lazy-choice where choices are evaluated lazily.
    """

    def __init__(self, get_choices: Callable[[], Sequence[str]], case_sensitive: bool = False):
        self._get_choices = get_choices
        self.case_sensitive = case_sensitive
        # Note: Purposely avoid super init.

    @cached_property
    def choices(self) -> Sequence[str]:  # type: ignore[override]
        return self._get_choices()



import re
from collections.abc import Sequence
from dataclasses import make_dataclass
from enum import Enum
from typing import Any, Optional, Union

from eth_abi import grammar
from eth_abi.abi import decode
from eth_abi.decoding import UnsignedIntegerDecoder
from eth_abi.encoding import UnsignedIntegerEncoder
from eth_abi.exceptions import DecodingError, InsufficientDataBytes
from eth_abi.registry import BaseEquals, registry
from eth_pydantic_types import HexBytes, HexStr
from eth_pydantic_types.utils import validate_bytes_size
from eth_utils import decode_hex
from ethpm_types.abi import ABIType, ConstructorABI, EventABI, EventABIType, MethodABI

from ape.logging import logger
from ape.utils.basemodel import ManagerAccessMixin

ARRAY_PATTERN = re.compile(r"[(*\w,? )]*\[\d*]")
NATSPEC_KEY_PATTERN = re.compile(r"(@\w+)")


class _ApeUnsignedIntegerDecoder(UnsignedIntegerDecoder):
    """
    This class exists because uint256 values when not-padded
    always cause issues, even with strict=False.
    It can be deleted if https://github.com/ethereum/eth-abi/pull/240
    merges.
    """

    def read_data_from_stream(self, stream):
        """
        Override to pad the value instead of raising an error.
        """
        data_byte_size: int = self.data_byte_size  # type: ignore
        data = stream.read(data_byte_size)

        if len(data) != data_byte_size:
            # Pad the value (instead of raising InsufficientBytesError).
            data = validate_bytes_size(data, 32)

        return data


registry.unregister("uint")
registry.register(
    BaseEquals("uint"),
    UnsignedIntegerEncoder,
    _ApeUnsignedIntegerDecoder,
    label="uint",
)


def is_array(abi_type: Union[str, ABIType]) -> bool:
    """
    Returns ``True`` if the given type is a probably an array.

    Args:
        abi_type (Union[str, ABIType]): The type to check.

    Returns:
        bool
    """

    return ARRAY_PATTERN.match(str(abi_type)) is not None


def returns_array(abi: MethodABI) -> bool:
    """
    Returns ``True`` if the given method ABI likely returns an array.

    Args:
        abi (MethodABI): An ABI method.

    Returns:
        bool
    """

    return _is_array_return(abi.outputs)


def _is_array_return(outputs: Sequence[ABIType]):
    return len(outputs) == 1 and is_array(outputs[0].type)


class StructParser:
    """
    A utility class responsible for parsing structs out of values.
    """

    def __init__(self, method_abi: Union[ConstructorABI, MethodABI, EventABI]):
        self.abi = method_abi

    @property
    def default_name(self) -> str:
        """
        The default struct return name for unnamed structs.
        This value is also used for named tuples where the tuple does not have a name
        (but each item in the tuple does have a name).
        """
        name = self.abi.name if isinstance(self.abi, MethodABI) else "constructor"
        return f"{name}_return"

    def encode_input(self, values: Union[list, tuple, dict]) -> Any:
        """
        Convert dicts and other objects to struct inputs.

        Args:
            values (Union[list, tuple]): A list of input values.

        Returns:
            Any: The same input values only decoded into structs when applicable.
        """
        return [self._encode(ipt, v) for ipt, v in zip(self.abi.inputs, values)]

    def decode_input(self, values: Union[Sequence, dict[str, Any]]) -> Any:
        return (
            self._decode(self.abi.inputs, values)
            if isinstance(self.abi, (EventABI, MethodABI))
            else None
        )

    def _encode(self, _type: ABIType, value: Any):
        if (
            _type.type == "tuple"
            and _type.components
            and all(m.name for m in _type.components)
            and not isinstance(value, tuple)
        ):
            if isinstance(value, dict):
                return tuple(
                    (
                        self._encode(m, value[m.name])
                        if isinstance(value[m.name], dict)
                        else value[m.name]
                    )
                    for m in _type.components
                )

            elif isinstance(value, (list, tuple)):
                # NOTE: Args must be passed in correct order.
                return tuple(value)

            else:
                arg = [getattr(value, m.name) for m in _type.components if m.name]
                return tuple(arg)

        elif (
            str(_type.type).startswith("tuple[")
            and isinstance(value, (list, tuple))
            and len(_type.components or []) > 0
        ):
            non_array_type_data = _type.model_dump()
            non_array_type_data["type"] = "tuple"
            non_array_type = ABIType(**non_array_type_data)
            return [self._encode(non_array_type, v) for v in value]

        return value

    def decode_output(self, values: Union[list, tuple]) -> Any:
        """
        Parse a list of output types and values into structs.
        Values are only altered when they are a struct.
        This method also handles structs within structs as well as arrays of structs.

        Args:
            values (Union[list, tuple]): A list of output values.

        Returns:
            Any: The same input values only decoded into structs when applicable.
        """

        return self._decode(self.abi.outputs, values) if isinstance(self.abi, MethodABI) else None

    def _decode(
        self,
        _types: Union[Sequence[ABIType]],
        values: Union[Sequence, dict[str, Any]],
    ):
        if is_struct(_types):
            return self._create_struct(_types[0], values)

        elif isinstance(values, (list, tuple)) and is_named_tuple(_types, values):
            # Handle tuples. NOTE: unnamed output structs appear as tuples with named members
            return create_struct(self.default_name, _types, values)

        return_values: list = []
        has_array_return = _is_array_return(_types)
        has_array_of_tuples_return = (
            has_array_return and len(_types) == 1 and "tuple" in _types[0].type
        )
        if has_array_return and not has_array_of_tuples_return:
            # Normal array
            return values

        elif has_array_of_tuples_return:
            item_type_str = str(_types[0].type).partition("[")[0]
            data = {
                **_types[0].model_dump(),
                "type": item_type_str,
                "internalType": item_type_str,
            }
            output_type = ABIType.model_validate(data)

            if isinstance(values, (list, tuple)) and not values[0]:
                # Only returned an empty list.
                return_values.append([])

            elif isinstance(values, (list, tuple)):
                for value in values[0]:
                    item = self._decode([output_type], [value])
                    return_values.append(item)

        else:
            for output_type, value in zip(_types, values):
                if isinstance(value, (tuple, list)):
                    item_type_str = str(output_type.type).partition("[")[0]
                    if item_type_str == "tuple":
                        # Either an array of structs or nested structs.
                        item_type_data = {
                            **output_type.model_dump(),
                            "type": item_type_str,
                            "internalType": item_type_str,
                        }
                        item_type = ABIType.model_validate(item_type_data)

                        if is_struct(output_type):
                            parsed_item = self._decode([item_type], [value])
                        else:
                            # Is array of structs.
                            parsed_item = [self._decode([item_type], [v]) for v in value]

                        # If it's an empty dynamic array of structs, replace `None` with empty list
                        output_raw_type = output_type.type
                        if (
                            isinstance(output_raw_type, str)
                            and output_raw_type.endswith("[]")
                            and parsed_item is None
                        ):
                            parsed_item = []

                    else:
                        parsed_item = [HexBytes(v) if isinstance(v, bytes) else v for v in value]

                    return_values.append(parsed_item)

                else:
                    return_values.append(value)

        return return_values

    def _create_struct(self, out_abi: ABIType, out_value: Any) -> Optional[Any]:
        if not out_abi.components or not out_value[0]:
            # Likely an empty tuple or not a struct.
            return None

        internal_type = out_abi.internal_type
        if out_abi.name == "" and internal_type and "struct " in internal_type:
            name = internal_type.replace("struct ", "").split(".")[-1]
        else:
            name = out_abi.name or self.default_name

        components = self._parse_components(out_abi.components, out_value[0])
        return create_struct(
            name,
            out_abi.components,
            components,
        )

    def _parse_components(self, components: list[ABIType], values) -> list:
        parsed_values = []
        for component, value in zip(components, values):
            if is_struct(component):
                new_value = self._create_struct(component, (value,))
                parsed_values.append(new_value)
            elif is_array(component.type) and "tuple" in component.type and component.components:
                new_value = [self._decode(component.components, v) for v in value]
                parsed_values.append(new_value)
            else:
                parsed_values.append(value)

        return parsed_values


def is_struct(outputs: Union[ABIType, Sequence[ABIType]]) -> bool:
    """
    Returns ``True`` if the given output is a struct.
    """

    outputs_seq = outputs if isinstance(outputs, (tuple, list)) else [outputs]
    return (
        len(outputs_seq) == 1
        and "[" not in outputs_seq[0].type
        and outputs_seq[0].components not in (None, [])
        and all(c.name != "" for c in outputs_seq[0].components or [])
    )


def is_named_tuple(outputs: Sequence[ABIType], output_values: Sequence) -> bool:
    """
    Returns ``True`` if the given output is a tuple where every item is named.
    """

    return all(o.name for o in outputs) and len(output_values) > 1


class Struct:
    """
    A class for contract return values using the struct data-structure.
    """

    def items(self) -> dict:
        """Override"""
        return {}

    def __setitem__(self, key, value):
        """Override"""


def create_struct(name: str, types: Sequence[ABIType], output_values: Sequence) -> Any:
    """
    Create a dataclass representing an ABI struct that can be used as inputs or outputs.
    The struct properties can be accessed via ``.`` notation, as keys in a dictionary, or
    numeric tuple access.

    **NOTE**: This method assumes you already know the values to give to the struct
    properties.

    Args:
        name (str): The name of the struct.
        types (list[ABIType]: The types of values in the struct.
        output_values (list[Any]): The struct property values.

    Returns:
        Any: The struct dataclass.
    """

    def get_item(struct, key) -> Any:
        # NOTE: Allow struct to function as a tuple and dict as well
        struct_values = tuple(getattr(struct, field) for field in struct.__dataclass_fields__)
        if isinstance(key, str):
            return dict(zip(struct.__dataclass_fields__, struct_values))[key]

        return struct_values[key]

    def set_item(struct, key, value):
        if isinstance(key, str):
            setattr(struct, key, value)
        else:
            struct_values = tuple(getattr(struct, field) for field in struct.__dataclass_fields__)
            field_to_set = struct_values[key]
            setattr(struct, field_to_set, value)

    def contains(struct, key):
        return key in struct.__dataclass_fields__

    def is_equal(struct, other) -> bool:
        if not hasattr(other, "__len__"):
            return NotImplemented

        _len = len(other)
        if _len != len(struct):
            return False

        if hasattr(other, "items"):
            # Struct or dictionary.
            for key, value in other.items():
                if key not in struct:
                    # Different object.
                    return False

                if struct[key] != value:
                    # Mismatched properties.
                    return False

            # Both objects represent the same struct.
            return True

        elif isinstance(other, (list, tuple)):
            # Allows comparing structs with sequence types.
            # NOTE: The order of the expected sequence matters!

            for itm1, itm2 in zip(struct.values(), other):
                if itm1 != itm2:
                    return False

            return True

        else:
            return NotImplemented

    def length(struct) -> int:
        return len(struct.__dataclass_fields__)

    def items(struct) -> list[tuple]:
        return [(k, struct[k]) for k, v in struct.__dataclass_fields__.items()]

    def values(struct) -> list[Any]:
        return [x[1] for x in struct.items()]

    def reduce(struct) -> tuple:
        return (create_struct, (name, types, output_values))

    # NOTE: Should never be "_{i}", but mypy complains and we need a unique value
    properties = [m.name or f"_{i}" for i, m in enumerate(types)]
    methods = {
        "__eq__": is_equal,
        "__getitem__": get_item,
        "__setitem__": set_item,
        "__contains__": contains,
        "__len__": length,
        "__reduce__": reduce,
        "items": items,
        "values": values,
    }

    if conflicts := [p for p in properties if p in methods]:
        conflicts_str = ", ".join(conflicts)
        logger.debug(
            "The following methods are unavailable on the struct "
            f"due to having the same name as a field: {conflicts_str}"
        )
        for conflict in conflicts:
            del methods[conflict]

    struct_def = make_dataclass(
        name,
        properties,
        namespace=methods,
        bases=(Struct,),  # We set a base class for subclass checking elsewhere.
    )

    return struct_def(*output_values)


def is_dynamic_sized_type(abi_type: Union[ABIType, str]) -> bool:
    parsed = grammar.parse(str(abi_type))
    return parsed.is_dynamic


class LogInputABICollection:
    def __init__(self, abi: EventABI):
        self.abi = abi
        self.topic_abi_types = [i for i in abi.inputs if i.indexed]
        self.data_abi_types: list[EventABIType] = [i for i in abi.inputs if not i.indexed]

        names = [i.name for i in abi.inputs]
        if len(set(names)) < len(names):
            raise ValueError("duplicate names found in log input", abi)

    @property
    def event_name(self):
        return self.abi.name

    def decode(
        self, topics: list[str], data: Union[str, bytes], use_hex_on_fail: bool = False
    ) -> dict:
        decoded = {}
        for abi, topic_value in zip(self.topic_abi_types, topics[1:]):
            # reference types as indexed arguments are written as a hash
            # https://docs.soliditylang.org/en/v0.8.15/contracts.html#events
            abi_type = "bytes32" if is_dynamic_sized_type(abi.type) else abi.canonical_type
            hex_value = decode_hex(topic_value)

            try:
                value = decode([abi_type], hex_value, strict=False)[0]
            except InsufficientDataBytes as err:
                if use_hex_on_fail:
                    if abi.name not in decoded:
                        # This allow logs to still be findable on the receipt.
                        decoded[abi.name] = hex_value

                    else:
                        raise DecodingError(str(err)) from err

            else:
                # The data was formatted correctly and we were able to decode logs.
                result = self.decode_value(abi_type, value)
                decoded[abi.name] = result

        data_abi_types = [abi.canonical_type for abi in self.data_abi_types]
        hex_data = decode_hex(data) if isinstance(data, str) else data
        try:
            data_values = decode(data_abi_types, hex_data)
        except InsufficientDataBytes as err:
            warning_message = f"Failed to decode log data '{self.event_name}'."

            # Try again with strict=False
            try:
                data_values = decode(data_abi_types, hex_data, strict=False)
            except Exception:
                # Even with strict=False, we failed to decode.
                # This should be a rare occasion, if it ever happens.
                logger.warn_from_exception(err, warning_message)
                if use_hex_on_fail:
                    for abi in self.data_abi_types:
                        if abi.name not in decoded:
                            # This allow logs to still be findable on the receipt.
                            decoded[abi.name] = hex_data

                else:
                    raise DecodingError(str(err)) from err

            else:
                # This happens when providers accidentally leave off trailing zeroes.
                warning_message = (
                    f"{warning_message} "
                    "However, we are able to get a value using decode(strict=False)"
                )
                logger.warn_from_exception(err, warning_message)
                for abi, value in zip(self.data_abi_types, data_values):
                    decoded[abi.name] = self.decode_value(abi.canonical_type, value)

        else:
            # The data was formatted correctly and we were able to decode logs.
            for abi, value in zip(self.data_abi_types, data_values):
                decoded[abi.name] = self.decode_value(abi.canonical_type, value)

        return decoded

    def decode_value(self, abi_type: str, value: Any) -> Any:
        if abi_type == "bytes32":
            return HexBytes(value)

        elif isinstance(value, (list, tuple)) and is_array(abi_type):
            sub_type = "[".join(abi_type.split("[")[:-1])
            return [self.decode_value(sub_type, v) for v in value]

        elif isinstance(value, (list, tuple)):
            parser = StructParser(self.abi)
            result = parser.decode_input([value])
            return result[0] if len(result) == 1 else result

        # NOTE: All the rest of the types are handled by the
        #  ecosystem API through the calling function.

        return value


def _enrich_natspec(natspec: str) -> str:
    # Ensure the natspec @-words are highlighted.
    replacement = r"[bright_red]\1[/]"
    return re.sub(NATSPEC_KEY_PATTERN, replacement, natspec)


def encode_topics(abi: EventABI, topics: Optional[dict[str, Any]] = None) -> list[HexStr]:
    """
    Encode the given topics using the given ABI. Useful for searching logs.

    Args:
        abi (EventABI): The event.
        topics (dict[str, Any] } None): Topic inputs to encode.

    Returns:
        list[str]: Encoded topics.
    """
    topics = topics or {}
    values = {}

    unnamed_iter = 0
    topic_inputs = {}
    for abi_input in abi.inputs:
        if not abi_input.indexed:
            continue

        if name := abi_input.name:
            topic_inputs[name] = abi_input
        else:
            topic_inputs[f"_{unnamed_iter}"] = abi_input

    for input_name, input_value in topics.items():
        if input_name not in topic_inputs:
            # Was trying to use data or is not part of search.
            continue

        input_type = topic_inputs[input_name].type
        if input_type == "address":
            convert = ManagerAccessMixin.conversion_manager.convert
            if isinstance(input_value, (list, tuple)):
                adjusted_value = []
                for addr in input_value:
                    if isinstance(addr, str):
                        adjusted_value.append(convert(addr))
                    else:
                        from ape.types import AddressType

                        adjusted_value.append(convert(addr, AddressType))

                input_value = adjusted_value

            elif not isinstance(input_value, str):
                from ape.types import AddressType

                input_value = convert(input_value, AddressType)

        values[input_name] = input_value

    return abi.encode_topics(values)  # type: ignore


class CalldataRepr(str, Enum):
    """
    Control the display of calldata in Ape.
    """

    abridged = "abridged"
    """
    Show the first and last 4-bytes of the calldata, enough
    to see the method ID.
    """

    full = "full"
    """
    Show the full calldata.
    """



import pytest

from ape.contracts.base import ContractCallHandler
from ape.exceptions import ContractNotFoundError
from ape.utils import ZERO_ADDRESS


def test_struct_input(
    call_handler_with_struct_input, struct_input_for_call, output_from_struct_input_call
):
    actual = call_handler_with_struct_input.encode_input(*struct_input_for_call)
    assert actual == output_from_struct_input_call


def test_call_contract_not_found(mocker, method_abi_with_struct_input, networks):
    (networks.ethereum.local.__dict__ or {}).pop("explorer", None)
    contract = mocker.MagicMock()
    contract.address = ZERO_ADDRESS
    contract.is_contract = False
    method = method_abi_with_struct_input
    handler = ContractCallHandler(contract=contract, abis=[method])
    expected = f"Failed to get contract type for address '{ZERO_ADDRESS}'."
    with pytest.raises(ContractNotFoundError, match=expected):
        handler()


def test_transact_contract_not_found(mocker, owner, method_abi_with_struct_input, networks):
    (networks.ethereum.local.__dict__ or {}).pop("explorer", None)
    contract = mocker.MagicMock()
    contract.address = ZERO_ADDRESS
    contract.is_contract = False
    method = method_abi_with_struct_input
    handler = ContractCallHandler(contract=contract, abis=[method])
    expected = rf"Failed to get contract type for address '{ZERO_ADDRESS}'\."
    with pytest.raises(ContractNotFoundError, match=expected):
        handler.transact(sender=owner)



import sys
from pathlib import Path
from typing import TYPE_CHECKING

import click

from ape.cli.arguments import contract_file_paths_argument
from ape.cli.options import (
    ape_cli_context,
    config_override_option,
    excluded_compilers_option,
    project_option,
)

if TYPE_CHECKING:
    from ethpm_types import ContractType


def _include_dependencies_callback(ctx, param, value):
    return value or ctx.obj.config_manager.get_config("compile").include_dependencies


@click.command(short_help="Compile select contract source files")
@ape_cli_context()
@project_option()
@contract_file_paths_argument()
@click.option(
    "-f",
    "--force",
    "use_cache",
    flag_value=False,
    default=True,
    is_flag=True,
    help="Force recompiling selected contracts",
)
@click.option(
    "-s",
    "--size",
    "display_size",
    default=False,
    is_flag=True,
    help="Show deployment bytecode size for all contracts",
)
@click.option(
    "--include-dependencies",
    is_flag=True,
    help="Also compile dependencies",
    callback=_include_dependencies_callback,
)
@excluded_compilers_option()
@config_override_option()
def cli(
    cli_ctx,
    project,
    file_paths: set[Path],
    use_cache: bool,
    display_size: bool,
    include_dependencies,
    excluded_compilers: list[str],
    config_override,
):
    """
    Compiles the manifest for this project and saves the results
    back to the manifest.

    Note that ape automatically recompiles any changed contracts each time
    a project is loaded. You do not have to manually trigger a recompile.
    """
    compiled = False
    errored = False

    if cfg := config_override:
        project.reconfigure(**cfg)

    if file_paths:
        contracts = {
            k: v.contract_type
            for k, v in project.load_contracts(
                *file_paths, use_cache=use_cache, excluded_compilers=excluded_compilers
            ).items()
        }
        cli_ctx.logger.success("'local project' compiled.")
        compiled = True
        if display_size:
            _display_byte_code_sizes(cli_ctx, contracts)

    if (include_dependencies or project.config.compile.include_dependencies) and len(
        project.dependencies
    ) > 0:
        for dependency in project.dependencies:
            if use_cache and dependency.compiled:
                continue

            # Even if compiling we failed, we at least tried,
            # and so we don't need to warn "Nothing to compile".
            compiled = True

            try:
                contract_types: dict[str, ContractType] = {
                    c.contract_type.name: c.contract_type
                    for c in dependency.compile(use_cache=use_cache, allow_install=True).values()
                }
            except Exception as err:
                msg = f"Dependency '{dependency.name}' not installed. Reason: {err}"
                cli_ctx.logger.error(msg)
                errored = True
                continue

            cli_ctx.logger.success(f"'{dependency.project.name}' compiled.")
            if display_size:
                _display_byte_code_sizes(cli_ctx, contract_types)

    if not compiled:
        from ape.utils.os import clean_path  # perf: lazy import

        folder = clean_path(project.contracts_folder)
        cli_ctx.logger.warning(f"Nothing to compile ({folder}).")

    if errored:
        # Ensure exit code.
        sys.exit(1)


def _display_byte_code_sizes(cli_ctx, contract_types: dict[str, "ContractType"]):
    # Display bytecode size for *all* contract types (not just ones we compiled)
    code_size = []
    for contract in contract_types.values():
        if not contract.deployment_bytecode:
            continue  # Skip if not bytecode to display

        bytecode = contract.deployment_bytecode.bytecode

        if bytecode:
            code_size.append((contract.name, len(bytecode) // 2))

    if not code_size:
        cli_ctx.logger.info("No contracts with bytecode to display")
        return

    click.echo()
    click.echo("============ Deployment Bytecode Sizes ============")
    indent = max(len(i[0]) for i in code_size)  # type: ignore
    for name, size in sorted(code_size, key=lambda k: k[1], reverse=True):
        pct = size / 24577
        # pct_color = color(next((i[1] for i in CODESIZE_COLORS if pct >= i[0]), ""))
        # TODO Get colors fixed for bytecode size output
        # click.echo(f"  {name:<{indent}}  -  {size:>6,}B  ({pct_color}{pct:.2%}{color})")
        click.echo(f"  {name:<{indent}}  -  {size:>6,}B  ({pct:.2%})")

    click.echo()
